# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from collections.abc import Iterable
from logging import getLogger
from os import makedirs
from pathlib import Path
from urllib.request import urlretrieve
from zipfile import ZipFile

import numpy as np
import torch
from torch.utils.data import random_split
from torchvision.datasets import ImageFolder
from torchvision.transforms import ToTensor
from tqdm import tqdm

from openfl.federated import PyTorchDataLoader
from openfl.utilities import validate_file_hash

logger = getLogger(__name__)


class PyTorchHistologyInMemory(PyTorchDataLoader):
    """PyTorch data loader for Histology dataset."""

    def __init__(self, data_path=None, batch_size=32, **kwargs):
        """Instantiate the data object.

        Args:
            data_path: The file path to the data. If None, initialize for model creation only.
            batch_size: The batch size of the data loader
            **kwargs: Additional arguments, passed to super init
             and load_mnist_shard
        """
        super().__init__(batch_size, random_seed=0, **kwargs)

        # Set Histology-specific default attributes
        self.feature_shape = [3, 150, 150]
        self.num_classes = 8

        # If data_path is None, this is being used for model initialization only
        if data_path is None:
            logger.info("Initializing dataloader for model creation only (no data loading)")
            return

        try:
            int(data_path)
        except ValueError:
            raise ValueError(
                "Expected `%s` to be representable as `int`, as it refers to the data shard " +
                "number used by the collaborator.",
                data_path
            )

        X_train, y_train, X_valid, y_valid = load_histology_shard(
            shard_num=int(data_path),
            feature_shape=self.feature_shape,
            num_classes=self.num_classes,
            **kwargs
        )
        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid

    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Returns:
            list: The shape of an example feature array [3, 150, 150] for Histology images.
        """
        return self.feature_shape

    def get_num_classes(self):
        """Returns the number of classes for classification tasks.

        Returns:
            int: The number of classes (8 for Histology dataset).
        """
        return self.num_classes


class HistologyDataset(ImageFolder):
    """Colorectal Histology Dataset."""

    URL = ('https://zenodo.org/record/53169/files/Kather_'
           'texture_2016_image_tiles_5000.zip?download=1')
    FILENAME = 'Kather_texture_2016_image_tiles_5000.zip'
    FOLDER_NAME = 'Kather_texture_2016_image_tiles_5000'
    ZIP_SHA384 = ('7d86abe1d04e68b77c055820c2a4c582a1d25d2983e38ab724e'
                  'ac75affce8b7cb2cbf5ba68848dcfd9d84005d87d6790')
    DEFAULT_PATH = Path.cwd().absolute() / 'data'

    def __init__(self, root: Path = DEFAULT_PATH, **kwargs) -> None:
        """Initialize."""
        makedirs(root, exist_ok=True)
        filepath = root / HistologyDataset.FILENAME
        if not filepath.is_file():
            self.pbar = tqdm(total=None)
            urlretrieve(HistologyDataset.URL, filepath, self.report_hook)  # nosec
            validate_file_hash(filepath, HistologyDataset.ZIP_SHA384)
            with ZipFile(filepath, 'r') as f:
                f.extractall(root)

        super(HistologyDataset, self).__init__(root / HistologyDataset.FOLDER_NAME, **kwargs)

    def report_hook(self, count, block_size, total_size):
        """Update progressbar."""
        if self.pbar.total is None and total_size:
            self.pbar.total = total_size
        progress_bytes = count * block_size
        self.pbar.update(progress_bytes - self.pbar.n)

    def __getitem__(self, index):
        """Allow getting items by slice index."""
        if isinstance(index, Iterable):
            return [super(HistologyDataset, self).__getitem__(i) for i in index]
        else:
            return super(HistologyDataset, self).__getitem__(index)


def one_hot(labels, classes):
    """
    One Hot encode a vector.

    Args:
        labels (list):  List of labels to onehot encode
        classes (int): Total number of categorical classes

    Returns:
        np.array: Matrix of one-hot encoded labels
    """
    return np.eye(classes)[labels]


def _load_raw_datashards(shard_num, collaborator_count, train_split_ratio=0.8):
    """
    Load the raw data by shard.

    Returns tuples of the dataset shard divided into training and validation.

    Args:
        shard_num (int): The shard number to use
        collaborator_count (int): The number of collaborators in the federation

    Returns:
        2 tuples: (image, label) of the training, validation dataset
    """
    dataset = HistologyDataset(transform=ToTensor())
    n_train = int(train_split_ratio * len(dataset))
    n_valid = len(dataset) - n_train
    ds_train, ds_val = random_split(
        dataset, lengths=[n_train, n_valid], generator=torch.manual_seed(0))

    # create the shards
    X_train, y_train = list(zip(*ds_train[shard_num::collaborator_count]))
    X_train, y_train = np.stack(X_train), np.array(y_train)

    X_valid, y_valid = list(zip(*ds_val[shard_num::collaborator_count]))
    X_valid, y_valid = np.stack(X_valid), np.array(y_valid)

    return (X_train, y_train), (X_valid, y_valid)


def load_histology_shard(shard_num, collaborator_count, feature_shape=None, num_classes=None,
                         categorical=False, channels_last=False, **kwargs):
    """
    Load the Histology dataset.

    Args:
        shard_num (int): The shard to use from the dataset
        collaborator_count (int): The number of collaborators in the federation
        feature_shape (list, optional): The shape of input features.
        num_classes (int, optional): Number of classes.
        categorical (bool): True = convert the labels to one-hot encoded
         vectors (Default = True)
        channels_last (bool): True = The input images have the channels
         last (Default = True)
        **kwargs: Additional parameters to pass to the function

    Returns:
        numpy.ndarray: The training data
        numpy.ndarray: The training labels
        numpy.ndarray: The validation data
        numpy.ndarray: The validation labels
    """
    img_rows, img_cols = feature_shape[1], feature_shape[2]

    (X_train, y_train), (X_valid, y_valid) = _load_raw_datashards(
        shard_num, collaborator_count)

    if channels_last:
        X_train = X_train.reshape(X_train.shape[0], img_rows, img_cols, 3)
        X_valid = X_valid.reshape(X_valid.shape[0], img_rows, img_cols, 3)
    else:
        X_train = X_train.reshape(X_train.shape[0], 3, img_rows, img_cols)
        X_valid = X_valid.reshape(X_valid.shape[0], 3, img_rows, img_cols)

    logger.info(f'Histology > X_train Shape : {X_train.shape}')
    logger.info(f'Histology > y_train Shape : {y_train.shape}')
    logger.info(f'Histology > Train Samples : {X_train.shape[0]}')
    logger.info(f'Histology > Valid Samples : {X_valid.shape[0]}')

    if categorical:
        # convert class vectors to binary class matrices
        y_train = one_hot(y_train, num_classes)
        y_valid = one_hot(y_valid, num_classes)

    return X_train, y_train, X_valid, y_valid



# Copyright (C) 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import xml.etree.ElementTree as ETree
import json
import os
import math
import hashlib


def xml_to_json(input_base_folder, subfolders, output_folder, verify_hash=1):

    if not os.path.exists(input_base_folder):
        raise SystemExit(f"The folder '{input_base_folder}' does not exist.")

    train_data = []
    test_data = []
    train_count, test_count = 0, 0

    if verify_hash == 1:
        expected_hash = ('9d645c469ba37eb9ec2e121ae6ac90fbebccfb91f2aff7f'
                         'faabc0531f2ede54ab4c91bea775922e5910b276340c040e8')
        verify_aggregated_hashes(input_base_folder, subfolders,
                                 expected_hash=expected_hash)

    for subfolder in subfolders:
        folder_path = os.path.join(input_base_folder, subfolder)
        if os.path.isdir(folder_path):
            xml_files = [f for f in os.listdir(folder_path) if f.endswith('.xml')]
            test_file_count = math.ceil(len(xml_files) * 0.01)

            # Process files for training data
            for xml_file in xml_files[:-test_file_count]:
                new_data, count = process_xml_file(folder_path, xml_file)
                train_data.extend(new_data)
                train_count += count

            # Process files for test data
            for xml_file in xml_files[-test_file_count:]:
                new_data, count = process_xml_file(folder_path, xml_file)
                test_data.extend(new_data)
                test_count += count
        else:
            raise SystemError(f"{folder_path} does not exist")

    # Save the data to JSON files
    save_json(train_data, os.path.join(output_folder, 'medquad_alpaca_train.json'))
    save_json(test_data, os.path.join(output_folder, 'medquad_alpaca_test.json'))

    # Write the counts to a text file
    with open(os.path.join(output_folder, 'data_counts.txt'), 'w') as f:
        f.write(f"Training data pairs: {train_count}\n")
        f.write(f"Test data pairs: {test_count}\n")

    print("Preprocessing complete")


def process_xml_file(folder, xml_file):
    xml_path = os.path.join(folder, xml_file)
    tree = ETree.parse(xml_path)
    root = tree.getroot()

    data = []
    count = 0
    for qa_pair in root.findall(".//QAPair"):
        question = qa_pair.find('Question').text
        answer = qa_pair.find('Answer').text

        if not question or not answer:
            continue

        question = question.strip()
        answer = answer.strip().replace('\n', ' ').replace('  ', ' ')

        json_obj = {
            "instruction": question,
            "input": "",
            "output": answer
        }
        data.append(json_obj)
        count += 1

    return data, count


def save_json(data, filename):
    with open(filename, 'w', encoding='utf-8') as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


def compute_hash(file_path, hash_name='sha384'):
    """Compute the hash of a single file using SHA-384."""
    hash_func = getattr(hashlib, hash_name)()
    with open(file_path, 'rb') as f:
        for chunk in iter(lambda: f.read(8192), b''):
            hash_func.update(chunk)
    return hash_func.hexdigest()


def verify_aggregated_hashes(input_base_folder, dir_list, expected_hash):
    """Verify the aggregated hash of all files against a single, hardcoded hash."""
    aggregated_hash_func = hashlib.sha384()

    for sub_directory in dir_list:
        directory = os.path.join(input_base_folder, sub_directory)
        if os.path.isdir(directory):
            for root, _, files in os.walk(directory):
                for file in files:
                    file_path = os.path.join(root, file)
                    file_hash = compute_hash(file_path)
                    aggregated_hash_func.update(file_hash.encode('utf-8'))
        else:
            raise SystemError(f"{directory} does not exist")

    # Compute the aggregated hash
    aggregated_hash = aggregated_hash_func.hexdigest()

    # Compare the aggregated hash with the expected, hardcoded hash
    if aggregated_hash != expected_hash:
        raise SystemError(
            "Verification failed. Downloaded hash doesn\'t match expected hash.")
    else:
        print("Verification passed")



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os

import psutil

from openfl.callbacks.callback import Callback

logger = logging.getLogger(__name__)


class MemoryProfiler(Callback):
    """Profile memory usage of the current process at the end of each round.

    Attributes:
        log_dir: If set, writes logs as lines of JSON.
    """

    def __init__(self, log_dir: str = "./logs/"):
        super().__init__()
        self.log_dir = None
        if log_dir:
            os.makedirs(log_dir, exist_ok=True)
            self.log_dir = log_dir

    def on_round_end(self, round_num: int, logs=None):
        origin = self.params["origin"]

        info = _get_memory_usage()
        info["round_number"] = round_num
        info["origin"] = origin

        logger.info(f"Round {round_num}: Memory usage: {info}")
        if self.log_dir:
            with open(os.path.join(self.log_dir, f"{origin}_memory_usage.txt"), "a") as f:
                f.write(json.dumps(info) + "\n")


def _get_memory_usage() -> dict:
    process = psutil.Process(os.getpid())
    virtual_memory = psutil.virtual_memory()
    swap_memory = psutil.swap_memory()
    info = {
        "process_memory": round(process.memory_info().rss / (1024**2), 2),
        "virtual_memory/total": round(virtual_memory.total / (1024**2), 2),
        "virtual_memory/available": round(virtual_memory.available / (1024**2), 2),
        "virtual_memory/percent": virtual_memory.percent,
        "virtual_memory/used": round(virtual_memory.used / (1024**2), 2),
        "virtual_memory/free": round(virtual_memory.free / (1024**2), 2),
        "virtual_memory/active": round(virtual_memory.active / (1024**2), 2),
        "virtual_memory/inactive": round(virtual_memory.inactive / (1024**2), 2),
        "virtual_memory/buffers": round(virtual_memory.buffers / (1024**2), 2),
        "virtual_memory/cached": round(virtual_memory.cached / (1024**2), 2),
        "virtual_memory/shared": round(virtual_memory.shared / (1024**2), 2),
        "swap_memory/total": round(swap_memory.total / (1024**2), 2),
        "swap_memory/used": round(swap_memory.used / (1024**2), 2),
        "swap_memory/free": round(swap_memory.free / (1024**2), 2),
        "swap_memory/percent": swap_memory.percent,
    }
    return info



# Copyright (C) 2020-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from openfl.federated import PyTorchDataLoader
from src.mnist_utils import load_mnist_shard


class PyTorchMNISTInMemory(PyTorchDataLoader):
    """PyTorch data loader for MNIST dataset."""

    def __init__(self, data_path=None, batch_size=32, **kwargs):
        """Instantiate the data object.

        Args:
            data_path: The file path to the data. If None, initialize for model creation only.
            batch_size: The batch size of the data loader
            **kwargs: Additional arguments, passed to super
             init and load_mnist_shard
        """
        super().__init__(batch_size, **kwargs)

        # Set MNIST-specific default attributes
        self.train_loader = None
        self.val_loader = None
        self.feature_shape = [1, 28, 28]
        self.num_classes = 10

        # If data_path is None, this is being used for model initialization only
        if data_path is None:
            return

        # TODO: We should be downloading the dataset shard into a directory
        # TODO: There needs to be a method to ask how many collaborators and
        #  what index/rank is this collaborator.
        # Then we have a way to automatically shard based on rank and size
        # of collaborator list.

        try:
            int(data_path)
        except ValueError:
            raise ValueError(
                f"Expected '{data_path}' to be representable as `int`, "
                "as it refers to the data shard number used by the collaborator."
            )

        X_train, y_train, X_valid, y_valid = load_mnist_shard(
            shard_num=int(data_path), num_classes=self.num_classes, **kwargs
        )
        self.X_train = X_train
        self.y_train = y_train
        self.train_loader = self.get_train_loader()

        self.X_valid = X_valid
        self.y_valid = y_valid
        self.val_loader = self.get_valid_loader()

    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Returns:
            list: The shape of an example feature array [1, 28, 28] for MNIST.
        """
        return self.feature_shape

    def get_num_classes(self):
        """Returns the number of classes for classification tasks.

        Returns:
            int: The number of classes (10 for MNIST).
        """
        return self.num_classes



# Copyright (C) 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

import torch
import torchvision

train_dataset = torchvision.datasets.MNIST(
    "./files/",
    train=True,
    download=True,
    transform=torchvision.transforms.Compose(
        [
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.1307,), (0.3081,)),
        ]
    ),
)

test_dataset = torchvision.datasets.MNIST(
    "./files/",
    train=False,
    download=True,
    transform=torchvision.transforms.Compose(
        [
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.1307,), (0.3081,)),
        ]
    ),
)


def bangalore_attrs(index, n_collaborators, batch_size, train_dataset, test_dataset):
    train = deepcopy(train_dataset)
    test = deepcopy(test_dataset)
    train.data = train_dataset.data[index::n_collaborators]
    train.targets = train_dataset.targets[index::n_collaborators]
    test.data = test_dataset.data[index::n_collaborators]
    test.targets = test_dataset.targets[index::n_collaborators]

    return {
        "train_loader": torch.utils.data.DataLoader(train, batch_size=batch_size, shuffle=True),
        "test_loader": torch.utils.data.DataLoader(test, batch_size=batch_size, shuffle=True),
    }



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
import logging
from pathlib import Path

from openfl.callbacks.callback import Callback
from openfl.federated import Plan
from openfl.protocols import utils
from openfl.utilities.dataloading import initialize_minimal_dataloader
from openfl.utilities.workspace import set_directory

logger = logging.getLogger(__name__)


class ModelCheckpoint(Callback):
    """Save the model in native format at the end of the experiment.

    This callback saves the model in native format at the end of the experiment.
    It uses the `TaskRunner` to load the model from a protobuf file and then saves
    it in native format.

    Args:
        best_state_path (str): Path to the best model state file.
        last_state_path (str): Path to the last model state file.

    """

    def __init__(self, best_state_path, last_state_path):
        super().__init__()
        self.best_state_path = best_state_path
        self.last_state_path = last_state_path

    def on_experiment_end(self, logs=None):
        task_runner, tensor_pipe = initialize_task_runner()

        for state_path in [self.best_state_path, self.last_state_path]:
            state_path = Path(state_path).resolve()

            model_protobuf = utils.load_proto(state_path)

            tensor_dict, _ = utils.deconstruct_model_proto(model_protobuf, tensor_pipe)

            task_runner.set_tensor_dict(tensor_dict, with_opt_vars=False)
            state_path_no_ext = state_path.with_suffix("")
            output_filepath = task_runner.save_native(state_path_no_ext)
            logger.info("Model saved in native format at: %s", output_filepath)


def initialize_task_runner(
    plan_config: str = "plan/plan.yaml",
    cols_config: str = "plan/cols.yaml",
    data_config: str = "plan/data.yaml",
):
    """
    Initialize TaskRunner and load it with provided model.pbuf.

    Args:
        plan_config (str): Federated learning plan.
        cols_config (str): Authorized collaborator list.
        data_config (str): The data set/shard configuration file.

    Returns:
        task_runner (instance): TaskRunner instance.
        tensor_pipe (instance): TensorPipe instance.
    """

    # Here we change cwd to the experiment workspace folder
    # because plan.yaml usually contains relative paths to components.
    workspace_path = Path(plan_config).resolve().parent.parent
    plan_config = Path(plan_config).resolve().relative_to(workspace_path)
    cols_config = Path(cols_config).resolve().relative_to(workspace_path)
    data_config = Path(data_config).resolve().relative_to(workspace_path)

    with set_directory(workspace_path):
        plan = Plan.parse(
            plan_config_path=plan_config,
            cols_config_path=cols_config,
            data_config_path=data_config,
            resolve=False,
        )
        data_loader = initialize_minimal_dataloader(plan)
        task_runner = plan.get_task_runner(data_loader=data_loader)
        tensor_pipe = plan.get_tensor_pipe()

    del task_runner.data_loader
    return task_runner, tensor_pipe



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ModelProto(_message.Message):
    __slots__ = ("tensors",)
    TENSORS_FIELD_NUMBER: _ClassVar[int]
    tensors: _containers.RepeatedCompositeFieldContainer[NamedTensor]
    def __init__(self, tensors: _Optional[_Iterable[_Union[NamedTensor, _Mapping]]] = ...) -> None: ...

class NamedTensor(_message.Message):
    __slots__ = ("name", "round_number", "lossless", "report", "tags", "transformer_metadata", "data_bytes")
    NAME_FIELD_NUMBER: _ClassVar[int]
    ROUND_NUMBER_FIELD_NUMBER: _ClassVar[int]
    LOSSLESS_FIELD_NUMBER: _ClassVar[int]
    REPORT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    TRANSFORMER_METADATA_FIELD_NUMBER: _ClassVar[int]
    DATA_BYTES_FIELD_NUMBER: _ClassVar[int]
    name: str
    round_number: int
    lossless: bool
    report: bool
    tags: _containers.RepeatedScalarFieldContainer[str]
    transformer_metadata: _containers.RepeatedCompositeFieldContainer[MetadataProto]
    data_bytes: bytes
    def __init__(self, name: _Optional[str] = ..., round_number: _Optional[int] = ..., lossless: bool = ..., report: bool = ..., tags: _Optional[_Iterable[str]] = ..., transformer_metadata: _Optional[_Iterable[_Union[MetadataProto, _Mapping]]] = ..., data_bytes: _Optional[bytes] = ...) -> None: ...

class MetadataProto(_message.Message):
    __slots__ = ("int_to_float", "int_list", "bool_list")
    class IntToFloatEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: float
        def __init__(self, key: _Optional[int] = ..., value: _Optional[float] = ...) -> None: ...
    INT_TO_FLOAT_FIELD_NUMBER: _ClassVar[int]
    INT_LIST_FIELD_NUMBER: _ClassVar[int]
    BOOL_LIST_FIELD_NUMBER: _ClassVar[int]
    int_to_float: _containers.ScalarMap[int, float]
    int_list: _containers.RepeatedScalarFieldContainer[int]
    bool_list: _containers.RepeatedScalarFieldContainer[bool]
    def __init__(self, int_to_float: _Optional[_Mapping[int, float]] = ..., int_list: _Optional[_Iterable[int]] = ..., bool_list: _Optional[_Iterable[bool]] = ...) -> None: ...

class DataStream(_message.Message):
    __slots__ = ("size", "npbytes")
    SIZE_FIELD_NUMBER: _ClassVar[int]
    NPBYTES_FIELD_NUMBER: _ClassVar[int]
    size: int
    npbytes: bytes
    def __init__(self, size: _Optional[int] = ..., npbytes: _Optional[bytes] = ...) -> None: ...



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.notebooktools package."""

from openfl.experimental.workflow.notebooktools.code_analyzer import CodeAnalyzer
from openfl.experimental.workflow.notebooktools.notebook_tools import NotebookTools



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""PyTorch FedProx optimizer module."""

import math

import torch
from torch.optim import Optimizer
from torch.optim.optimizer import required


class FedProxOptimizer(Optimizer):
    """FedProx optimizer.

    Implements the FedProx optimization algorithm using PyTorch.
    FedProx is a federated learning optimization algorithm designed to handle
    non-IID data.
    It introduces a proximal term to the federated averaging algorithm to
    reduce the impact of devices with outlying updates.

    Paper: https://arxiv.org/pdf/1812.06127.pdf

    Attributes:
        params: Parameters to be stored for optimization.
        lr: Learning rate.
        mu: Proximal term coefficient.
        momentum: Momentum factor.
        dampening: Dampening for momentum.
        weight_decay: Weight decay (L2 penalty).
        nesterov: Enables Nesterov momentum.
    """

    def __init__(
        self,
        params,
        lr=required,
        mu=0.0,
        momentum=0,
        dampening=0,
        weight_decay=0,
        nesterov=False,
    ):
        """
        Initialize the FedProx optimizer.

        Args:
            params: Parameters to be stored for optimization.
            lr: Learning rate.
            mu: Proximal term coefficient. Defaults to 0.0.
            momentum: Momentum factor. Defaults to 0.
            dampening: Dampening for momentum. Defaults to 0.
            weight_decay: Weight decay (L2 penalty). Defaults to 0.
            nesterov: Enables Nesterov momentum. Defaults to False

        Raises:
            ValueError: If momentum is less than 0.
            ValueError: If learning rate is less than 0.
            ValueError: If weight decay is less than 0.
            ValueError: If mu is less than 0.
        """
        if momentum < 0.0:
            raise ValueError(f"Invalid momentum value: {momentum}")
        if lr is not required and lr < 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if weight_decay < 0.0:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")
        if mu < 0.0:
            raise ValueError(f"Invalid mu value: {mu}")
        defaults = {
            "dampening": dampening,
            "lr": lr,
            "momentum": momentum,
            "mu": mu,
            "nesterov": nesterov,
            "weight_decay": weight_decay,
        }

        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")

        super().__init__(params, defaults)

    def __setstate__(self, state):
        """
        Set optimizer state.

        Args:
            state: State dictionary.
        """
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("nesterov", False)

    @torch.no_grad()
    def step(self, closure=None):
        """Perform a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.

        Returns:
            Loss value if closure is provided. None otherwise.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group["weight_decay"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            nesterov = group["nesterov"]
            mu = group["mu"]
            w_old = group["w_old"]
            for p, w_old_p in zip(group["params"], w_old):
                if p.grad is None:
                    continue
                d_p = p.grad
                if weight_decay != 0:
                    d_p = d_p.add(p, alpha=weight_decay)
                if momentum != 0:
                    param_state = self.state[p]
                    if "momentum_buffer" not in param_state:
                        buf = param_state["momentum_buffer"] = torch.clone(d_p).detach()
                    else:
                        buf = param_state["momentum_buffer"]
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf
                if w_old is not None:
                    d_p.add_(p - w_old_p, alpha=mu)
                p.add_(d_p, alpha=-group["lr"])

        return loss

    def set_old_weights(self, old_weights):
        """Set the global weights parameter to `old_weights` value.

        Args:
            old_weights: The old weights to be set.
        """
        for param_group in self.param_groups:
            param_group["w_old"] = old_weights


class FedProxAdam(Optimizer):
    """FedProxAdam optimizer.

    Implements the FedProx optimization algorithm with Adam optimizer.

    Attributes:
        params: Parameters to be stored for optimization.
        mu: Proximal term coefficient.
        lr: Learning rate.
        betas: Coefficients used for computing running averages of gradient
            and its square.
        eps: Value for computational stability.
        weight_decay: Weight decay (L2 penalty).
        amsgrad: Whether to use the AMSGrad variant of this algorithm.
    """

    def __init__(
        self,
        params,
        mu=0,
        lr=1e-3,
        betas=(0.9, 0.999),
        eps=1e-8,
        weight_decay=0,
        amsgrad=False,
    ):
        """
        Args:
            params: Parameters to be stored for optimization.
            mu: Proximal term coefficient. Defaults to 0.
            lr: Learning rate. Defaults to 1e-3.
            betas: Coefficients used for computing running averages of
                gradient and its square. Defaults to (0.9, 0.999).
            eps: Value for computational stability. Defaults to 1e-8.
            weight_decay: Weight decay (L2 penalty). Defaults to 0.
            amsgrad: Whether to use the AMSGrad variant of this algorithm.
                Defaults to False.

        Raises:
            ValueError: If learning rate is less than 0.
            ValueError: If betas[0] is not in [0, 1).
            ValueError: If betas[1] is not in [0, 1).
            ValueError: If weight decay is less than 0.
            ValueError: If mu is less than 0.
        """
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= eps:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        if not 0.0 <= weight_decay:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")
        if mu < 0.0:
            raise ValueError(f"Invalid mu value: {mu}")
        defaults = {
            "lr": lr,
            "betas": betas,
            "eps": eps,
            "weight_decay": weight_decay,
            "amsgrad": amsgrad,
            "mu": mu,
        }
        super().__init__(params, defaults)

    def __setstate__(self, state):
        """Set optimizer state."""
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsgrad", False)

    def set_old_weights(self, old_weights):
        """Set the global weights parameter to `old_weights` value.

        Args:
            old_weights: The old weights to be set.
        """
        for param_group in self.param_groups:
            param_group["w_old"] = old_weights

    @torch.no_grad()
    def step(self, closure=None):
        """Perform a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.

        Returns:
            Loss value if closure is provided. None otherwise.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            exp_avg_sqs = []
            max_exp_avg_sqs = []
            state_steps = []

            for p in group["params"]:
                if p.grad is not None:
                    params_with_grad.append(p)
                    if p.grad.is_sparse:
                        raise RuntimeError(
                            "Adam does not support sparse gradients, "
                            "please consider SparseAdam instead"
                        )
                    grads.append(p.grad)

                    state = self.state[p]
                    # Lazy state initialization
                    if len(state) == 0:
                        state["step"] = 0
                        # Exponential moving average of gradient values
                        state["exp_avg"] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        # Exponential moving average of squared gradient values
                        state["exp_avg_sq"] = torch.zeros_like(
                            p, memory_format=torch.preserve_format
                        )
                        if group["amsgrad"]:
                            # Maintains max of all exp. moving avg. of sq. grad. values
                            state["max_exp_avg_sq"] = torch.zeros_like(
                                p, memory_format=torch.preserve_format
                            )

                    exp_avgs.append(state["exp_avg"])
                    exp_avg_sqs.append(state["exp_avg_sq"])

                    if group["amsgrad"]:
                        max_exp_avg_sqs.append(state["max_exp_avg_sq"])

                    # update the steps for each param group update
                    state["step"] += 1
                    # record the step after step update
                    state_steps.append(state["step"])

            beta1, beta2 = group["betas"]
            self.adam(
                params_with_grad,
                grads,
                exp_avgs,
                exp_avg_sqs,
                max_exp_avg_sqs,
                state_steps,
                group["amsgrad"],
                beta1,
                beta2,
                group["lr"],
                group["weight_decay"],
                group["eps"],
                group["mu"],
                group["w_old"],
            )
        return loss

    def adam(
        self,
        params,
        grads,
        exp_avgs,
        exp_avg_sqs,
        max_exp_avg_sqs,
        state_steps,
        amsgrad,
        beta1: float,
        beta2: float,
        lr: float,
        weight_decay: float,
        eps: float,
        mu: float,
        w_old,
    ):
        """
        Update optimizer parameters.

        Args:
            params: Parameters to be stored for optimization.
            grads: Gradients.
            exp_avgs: Exponential moving average of gradient values.
            exp_avg_sqs: Exponential moving average of squared gradient values.
            max_exp_avg_sqs: Maintains max of all exp. moving avg. of sq. grad. values.
            state_steps: Steps for each param group update.
            amsgrad: Whether to use the AMSGrad variant of this algorithm.
            beta1 (float): Coefficient used for computing running averages of
                gradient.
            beta2 (float): Coefficient used for computing running averages of
                squared gradient.
            lr (float): Learning rate.
            weight_decay (float): Weight decay (L2 penalty).
            eps (float): Value for computational stability.
            mu (float): Proximal term coefficient.
            w_old: The old weights.
        """
        for i, param in enumerate(params):
            w_old_p = w_old[i]
            grad = grads[i]
            grad.add_(param - w_old_p, alpha=mu)
            exp_avg = exp_avgs[i]
            exp_avg_sq = exp_avg_sqs[i]
            step = state_steps[i]

            bias_correction1 = 1 - beta1**step
            bias_correction2 = 1 - beta2**step

            if weight_decay != 0:
                grad = grad.add(param, alpha=weight_decay)

            # Decay the first and second moment running average coefficient
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
            if amsgrad:
                # Maintains the maximum of all 2nd moment running avg. till now
                torch.maximum(max_exp_avg_sqs[i], exp_avg_sq, out=max_exp_avg_sqs[i])
                # Use the max. for normalizing running avg. of gradient
                denom = (max_exp_avg_sqs[i].sqrt() / math.sqrt(bias_correction2)).add_(eps)
            else:
                denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(eps)

            step_size = lr / bias_correction1

            param.addcdiv_(exp_avg, denom, value=-step_size)



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from openfl.utilities.secagg.crypto import (
    calculate_mask,
    calculate_shared_mask,
    create_ciphertext,
    decipher_ciphertext,
    pseudo_random_generator,
)
from openfl.utilities.secagg.key import generate_agreed_key, generate_key_pair
from openfl.utilities.secagg.shamir import create_secret_shares, reconstruct_secret



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.utilities.metaflow_utils module."""

from __future__ import annotations

import ast
import fcntl
import hashlib
import os
import tempfile
from datetime import datetime
from pathlib import Path

# getsource only used to determine structure of FlowGraph
from typing import TYPE_CHECKING

import cloudpickle as pickle
import ray
from dill.source import getsource  # nosec
from metaflow.datastore import DATASTORES, FlowDataStore
from metaflow.datastore.exceptions import DataException, UnpicklableArtifactException
from metaflow.datastore.task_datastore import TaskDataStore, only_if_not_done, require_mode
from metaflow.graph import DAGNode, FlowGraph, StepVisitor, deindent_docstring
from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.mflog import RUNTIME_LOG_SOURCE
from metaflow.plugins import LocalMetadataProvider
from metaflow.runtime import MAX_LOG_SIZE, TruncatedBuffer, mflog_msg
from metaflow.task import MetaDatum

if TYPE_CHECKING:
    from openfl.experimental.workflow.interface import FLSpec

import base64
import json
import uuid
from io import StringIO
from typing import Any, Generator, Type

from metaflow import __version__ as mf_version
from metaflow.plugins.cards.card_modules.basic import (
    CSS_PATH,
    JS_PATH,
    RENDER_TEMPLATE_PATH,
    DagComponent,
    DefaultCard,
    PageComponent,
    SectionComponent,
    TaskInfoComponent,
    read_file,
    transform_flow_graph,
)


class SystemMutex:
    """Provides a system-wide mutex that locks a file until the lock is
    released."""

    def __init__(self, name):
        """Initializes the SystemMutex with the provided name.

        Args:
            name (str): The name of the mutex.
        """
        self.name = name

    def __enter__(self):
        lock_id = hashlib.new("sha256", self.name.encode("utf8"), usedforsecurity=False).hexdigest()  # nosec
        # Using SHA-256 to address security warning
        self.fp = open(os.path.join(tempfile.mkdtemp(), f".lock-{lock_id}.lck"), "wb")
        fcntl.flock(self.fp.fileno(), fcntl.LOCK_EX)

    def __exit__(self, _type, value, tb):
        fcntl.flock(self.fp.fileno(), fcntl.LOCK_UN)
        self.fp.close()


class Flow:
    """A mock class representing a flow for Metaflow's internal use."""

    def __init__(self, name):
        """Mock flow for metaflow internals.

        Args:
            name (str): The name of the flow.
        """
        self.name = name


@ray.remote
class Counter:
    def __init__(self):
        """Initializes the Counter with value set to 0."""
        self.value = 0

    def increment(self):
        """Increments the counter by 1.

        Returns:
            int: The incremented value of the counter.
        """
        self.value += 1
        return self.value

    def get_counter(self):
        """Retrieves the current value of the counter.

        Returns:
            int: The current value of the counter.
        """
        return self.value


class DAGnode(DAGNode):
    """A custom DAGNode class for the Metaflow graph.

    Attributes:
        name (str): The name of the DAGNode.
        func_lineno (int): The line number of the function in the source code.
        decorators (list): The decorators applied to the function.
        doc (str): The docstring of the function.
        parallel_step (bool): A flag indicating if the step is parallelized.
    """

    def __init__(self, func_ast, decos, doc):
        """Initializes the DAGNode with the provided function AST, decorators,
        and docstring.

        Args:
            func_ast (ast.FunctionDef): The function's abstract syntax tree.
            decos (list): The decorators applied to the function.
            doc (str): The docstring of the function.
        """
        self.name = func_ast.name
        self.func_lineno = func_ast.lineno
        self.decorators = decos
        self.doc = deindent_docstring(doc)
        self.parallel_step = any(getattr(deco, "IS_PARALLEL", False) for deco in decos)

        # these attributes are populated by _parse
        self.tail_next_lineno = 0
        self.type = None
        self.out_funcs = []
        self.has_tail_next = False
        self.invalid_tail_next = False
        self.num_args = 0
        self.foreach_param = None
        self.num_parallel = 0
        self.parallel_foreach = False
        self._parse(func_ast)

        # these attributes are populated by _traverse_graph
        self.in_funcs = set()
        self.split_parents = []
        self.matching_join = None
        # these attributes are populated by _postprocess
        self.is_inside_foreach = False

    def _parse(self, func_ast):
        self.num_args = len(func_ast.args.args)
        tail = func_ast.body[-1]

        # End doesn't need a transition
        if self.name == "end":
            self.type = "end"
            return

        # Ensure that the tail is an expression
        if not isinstance(tail, ast.Expr):
            return

        # Determine the type of self.next transition
        try:
            if not self._expr_str(tail.value.func) == "self.next":
                return

            self.has_tail_next = True
            self.invalid_tail_next = True
            self.tail_next_lineno = tail.lineno
            self.out_funcs = [e.attr for e in tail.value.args]

            keywords = {k.arg: getattr(k.value, "s", None) for k in tail.value.keywords}

            # Determine the type based on keywords and out_funcs
            self._determine_type(keywords)
        except AttributeError:
            return

    def _determine_type(self, keywords):
        """Determine the type of the function based on keywords and out_funcs."""
        if len(keywords) in {1, 2}:
            if "foreach" in keywords:
                self._set_foreach_type(keywords["foreach"])
            elif "num_parallel" in keywords:
                self._set_parallel_foreach_type(keywords["num_parallel"])
            else:
                self.type = "linear"
        elif len(keywords) == 0:
            self._set_type_based_on_out_funcs()

    def _set_foreach_type(self, foreach_param):
        """Set the type to 'foreach'."""
        self.type = "foreach"
        if len(self.out_funcs) == 1:
            self.foreach_param = foreach_param
            self.invalid_tail_next = False

    def _set_parallel_foreach_type(self, num_parallel):
        """Set the type to 'parallel foreach'."""
        self.type = "foreach"
        self.parallel_foreach = True
        if len(self.out_funcs) == 1:
            self.num_parallel = num_parallel
            self.invalid_tail_next = False

    def _set_type_based_on_out_funcs(self):
        """Set the type based on the number of out_funcs."""
        if len(self.out_funcs) > 1:
            self.type = "split"
            self.invalid_tail_next = False
        elif len(self.out_funcs) == 1:
            if self.name == "start":
                self.type = "start"
            elif self.num_args > 1:
                self.type = "join"
            else:
                self.type = "linear"
            self.invalid_tail_next = False


class StepVisitor(StepVisitor):
    """A custom StepVisitor class for visiting the steps in a Metaflow
    graph."""

    def __init__(self, nodes, flow):
        """Initializes the StepVisitor with the provided nodes and flow.

        Args:
            nodes (dict): The nodes in the graph.
            flow (Flow): The flow object.
        """
        super().__init__(nodes, flow)

    def visit_FunctionDef(self, node):  # NOQA: N802
        """Visits a FunctionDef node in the flow and adds it to the nodes
        dictionary if it's a step.

        Args:
            node (ast.FunctionDef): The function definition node to visit.
        """
        func = getattr(self.flow, node.name)
        if hasattr(func, "is_step"):
            self.nodes[node.name] = DAGnode(node, func.decorators, func.__doc__)


class FlowGraph(FlowGraph):
    """A custom FlowGraph class for representing a Metaflow graph."""

    def __init__(self, flow):
        """Initializes the FlowGraph with the provided flow.

        Args:
            flow (Flow): The flow object.
        """
        self.name = flow.__name__
        self.nodes = self._create_nodes(flow)
        self.doc = deindent_docstring(flow.__doc__)
        self._traverse_graph()
        self._postprocess()

    def _create_nodes(self, flow):
        """Creates nodes for the flow graph by parsing the source code of the
        flow's module.

        Args:
            flow (Flow): The flow object.

        Returns:
            nodes (dict): A dictionary of nodes in the graph.
        """
        module = __import__(flow.__module__)
        tree = ast.parse(getsource(module)).body
        root = [n for n in tree if isinstance(n, ast.ClassDef) and n.name == self.name][0]
        nodes = {}
        StepVisitor(nodes, flow).visit(root)
        return nodes


class TaskDataStore(TaskDataStore):
    """A custom TaskDataStore class for storing task data in Metaflow."""

    def __init__(
        self,
        flow_datastore,
        run_id,
        step_name,
        task_id,
        attempt=None,
        data_metadata=None,
        mode="r",
        allow_not_done=False,
    ):
        """Initializes the TaskDataStore with the provided parameters.

        Args:
            flow_datastore (FlowDataStore): The flow datastore.
            run_id (str): The run id.
            step_name (str): The step name.
            task_id (str): The task id.
            attempt (int, optional): The attempt number. Defaults to None.
            data_metadata (DataMetadata, optional): The data metadata.
                Defaults to None.
            mode (str, optional): The mode (read 'r' or write 'w'). Defaults
                to 'r'.
            allow_not_done (bool, optional): A flag indicating whether to
                allow tasks that are not done. Defaults to False.
        """
        super().__init__(
            flow_datastore,
            run_id,
            step_name,
            task_id,
            attempt,
            data_metadata,
            mode,
            allow_not_done,
        )

    @only_if_not_done
    @require_mode("w")
    def save_artifacts(self, artifacts_iter, force_v4=False, len_hint=0):
        """Saves Metaflow Artifacts (Python objects) to the datastore and
        stores any relevant metadata needed to retrieve them.

        Typically, objects are pickled but the datastore may perform any
        operation that it deems necessary. You should only access artifacts
        using load_artifacts

        This method requires mode 'w'.

        Args:
            artifacts_iter (Iterator[(string, object)]): Iterator over the
                human-readable name of the object to save and the object
                itself.
            force_v4 (Union[bool, Dict[string -> boolean]], optional): Indicates
                whether the artifact should be pickled using the v4 version of
                pickle. If a single boolean, applies to all artifacts. If a
                dictionary, applies to the object named only. Defaults to False if
                not present or not specified.
            len_hint (int, optional): Estimated number of items in artifacts_iter.
                Defaults to 0.
        """
        artifact_names = []

        def pickle_object(name, obj, do_v4):
            if do_v4:
                return _pickle_with_protocol(name, obj, 4, "gzip+pickle-v4")
            else:
                try:
                    return _pickle_with_protocol(name, obj, 2, "gzip+pickle-v2")
                except (SystemError, OverflowError):
                    return _pickle_with_protocol(name, obj, 4, "gzip+pickle-v4")

        def _pickle_with_protocol(name, obj, protocol, encode_type):
            if encode_type not in self._encodings:
                raise DataException(
                    f"Artifact {name} requires a serialization encoding that "
                    "requires Python 3.4 or newer."
                )
            try:
                return pickle.dumps(obj, protocol=protocol), encode_type
            except TypeError:
                raise UnpicklableArtifactException(name)

        def pickle_iter():
            for name, obj in artifacts_iter:
                do_v4 = force_v4 if isinstance(force_v4, bool) else force_v4.get(name, False)
                blob, encode_type = pickle_object(name, obj, do_v4)
                self._info[name] = {
                    "size": len(blob),
                    "type": str(type(obj)),
                    "encoding": encode_type,
                }
                artifact_names.append(name)
                yield blob

        # Use the content-addressed store to store all artifacts
        save_result = self._ca_store.save_blobs(pickle_iter(), len_hint=len_hint)
        for name, result in zip(artifact_names, save_result):
            self._objects[name] = result.key


class FlowDataStore(FlowDataStore):
    """A custom FlowDataStore class for storing flow data in Metaflow."""

    def __init__(
        self,
        flow_name,
        environment,
        metadata=None,
        event_logger=None,
        monitor=None,
        storage_impl=None,
        ds_root=None,
    ):
        """Initializes the FlowDataStore with the provided parameters.

        Args:
            flow_name (str): The name of the flow.
            environment (MetaflowEnvironment): The Metaflow environment.
            metadata (MetadataProvider, optional): The metadata provider.
                Defaults to None.
            event_logger (EventLogger, optional): The event logger. Defaults
                to None.
            monitor (Monitor, optional): The monitor. Defaults to None.
            storage_impl (DataStore, optional): The storage implementation.
                Defaults to None.
            ds_root (str, optional): The root of the datastore. Defaults to
                None.
        """

        super().__init__(
            flow_name,
            environment,
            metadata,
            event_logger,
            monitor,
            storage_impl,
            ds_root,
        )

    def get_task_datastore(
        self,
        run_id,
        step_name,
        task_id,
        attempt=None,
        data_metadata=None,
        mode="r",
        allow_not_done=False,
    ):
        """Returns a TaskDataStore for the specified task.

        Args:
            run_id (str): The run id.
            step_name (str): The step name.
            task_id (str): The task id.
            attempt (int, optional): The attempt number. Defaults to None.
            data_metadata (DataMetadata, optional): The data metadata.
                Defaults to None.
            mode (str, optional): The mode (read 'r' or write 'w'). Defaults
                to 'r'.
            allow_not_done (bool, optional): A flag indicating whether to
                allow tasks that are not done. Defaults to False.

        Returns:
            TaskDataStore: A TaskDataStore for the specified task.
        """
        return TaskDataStore(
            self,
            run_id,
            step_name,
            task_id,
            attempt=attempt,
            data_metadata=data_metadata,
            mode=mode,
            allow_not_done=allow_not_done,
        )


class MetaflowInterface:
    """A wrapper class for Metaflow's tooling, modified to work with the
    workflow interface."""

    def __init__(self, flow: Type[FLSpec], backend: str = "ray"):
        """Wrapper class for the metaflow tooling modified to work with the
        workflow interface. Keeps track of the current flow run, tasks, and
        data artifacts.

        Args:
            flow (Type[FLSpec]): The current flow that will be serialized /
                tracked using metaflow tooling.
            backend (str, optional): The backend selected by the runtime.
                Permitted selections are 'ray' and 'single_process'. Defaults
                to 'ray'.
        """
        self.backend = backend
        self.flow_name = flow.__name__
        if backend == "ray":
            self.counter = Counter.remote()
        else:
            self.counter = 0

    def create_run(self) -> int:
        """Creates a run for the current flow using metaflow internal
        functions.

        Args:
            None

        Returns:
            run_id [int]
        """
        flow = Flow(self.flow_name)
        env = MetaflowEnvironment(self.flow_name)
        env.get_environment_info()
        self.local_metadata = LocalMetadataProvider(env, flow, None, None)
        self.run_id = self.local_metadata.new_run_id()
        self.flow_datastore = FlowDataStore(
            self.flow_name,
            env,
            metadata=self.local_metadata,
            storage_impl=DATASTORES["local"],
            ds_root=f"{Path.home()}/.metaflow",
        )
        return self.run_id

    def create_task(self, task_name: str) -> int:
        """Creates a task for the current run. The generated task_id is unique
        for each task and can be recalled later with the metaflow client.

        Args:
            task_name (str): The name of the new task.

        Returns:
            task_id [int]
        """
        with SystemMutex("critical_section"):
            if self.backend == "ray":
                task_id = ray.get(self.counter.get_counter.remote())
                self.local_metadata._task_id_seq = task_id
                self.local_metadata.new_task_id(self.run_id, task_name)
                return ray.get(self.counter.increment.remote())
            else:
                # Keeping single_process in critical_section
                # because gRPC calls may cause problems.
                task_id = self.counter
                self.local_metadata._task_id_seq = task_id
                self.local_metadata.new_task_id(self.run_id, task_name)
                self.counter += 1
                return self.counter

    def save_artifacts(
        self,
        data_pairs: Generator[str, Any],
        task_name: str,
        task_id: int,
        buffer_out: Type[StringIO],
        buffer_err: Type[StringIO],
    ) -> None:
        """Use metaflow task datastore to save flow attributes, stdout, and
        stderr for a specific task (identified by the task_name + task_id).

        Args:
            data_pairs (Generator[str, Any]): Generator that returns the name
                of the attribute, and it's corresponding object.
            task_name (str): The name of the task for which an artifact is
                being saved.
            task_id (int): A unique id (within the flow) that will be used to
                recover these data artifacts by the metaflow client.
            buffer_out (StringIO): StringIO buffer containing stdout.
            buffer_err (StringIO): StringIO buffer containing stderr.
        """
        task_datastore = self.flow_datastore.get_task_datastore(
            self.run_id, task_name, str(task_id), attempt=0, mode="w"
        )
        task_datastore.init_task()
        task_datastore.save_artifacts(data_pairs)

        # Register metadata for task
        retry_count = 0
        metadata_tags = [f"attempt_id:{retry_count}"]
        self.local_metadata.register_metadata(
            self.run_id,
            task_name,
            str(task_id),
            [
                MetaDatum(
                    field="attempt",
                    value=str(retry_count),
                    type="attempt",
                    tags=metadata_tags,
                ),
                MetaDatum(
                    field="origin-run-id",
                    value=str(0),
                    type="origin-run-id",
                    tags=metadata_tags,
                ),
                MetaDatum(
                    field="ds-type",
                    value=self.flow_datastore.TYPE,
                    type="ds-type",
                    tags=metadata_tags,
                ),
                MetaDatum(
                    field="ds-root",
                    value=self.flow_datastore.datastore_root,
                    type="ds-root",
                    tags=metadata_tags,
                ),
            ],
        )

        self.emit_log(buffer_out, buffer_err, task_datastore)

        task_datastore.done()

    def load_artifacts(self, artifact_names, task_name, task_id):
        """Loads flow attributes from Metaflow's task datastore.

        Args:
            artifact_names (list): The names of the artifacts to load.
            task_name (str): The name of the task from which to load artifacts.
            task_id (int): The id of the task from which to load artifacts.

        Returns:
            dict: A dictionary of loaded artifacts.
        """
        task_datastore = self.flow_datastore.get_task_datastore(
            self.run_id, task_name, str(task_id), attempt=0, mode="r"
        )
        return task_datastore.load_artifacts(artifact_names)

    def emit_log(
        self,
        msgbuffer_out: Type[StringIO],
        msgbuffer_err: Type[StringIO],
        task_datastore: Type[TaskDataStore],
        system_msg: bool = False,
    ) -> None:
        """Writes stdout and stderr to Metaflow's TaskDatastore.

        Args:
            msgbuffer_out (StringIO): A StringIO buffer containing stdout.
            msgbuffer_err (StringIO): A StringIO buffer containing stderr.
            task_datastore (TaskDataStore): A Metaflow TaskDataStore instance.
            system_msg (bool, optional): A flag indicating whether the message
                is a system message. Defaults to False.
        """
        stdout_buffer = TruncatedBuffer("stdout", MAX_LOG_SIZE)
        stderr_buffer = TruncatedBuffer("stderr", MAX_LOG_SIZE)

        for std_output in msgbuffer_out.readlines():
            timestamp = datetime.utcnow()
            stdout_buffer.write(mflog_msg(std_output, now=timestamp), system_msg=system_msg)

        for std_error in msgbuffer_err.readlines():
            timestamp = datetime.utcnow()
            stderr_buffer.write(mflog_msg(std_error, now=timestamp), system_msg=system_msg)

        task_datastore.save_logs(
            RUNTIME_LOG_SOURCE,
            {
                "stdout": stdout_buffer.get_buffer(),
                "stderr": stderr_buffer.get_buffer(),
            },
        )


class DefaultCard(DefaultCard):
    """A custom DefaultCard class for Metaflow.

    Attributes:
        ALLOW_USER_COMPONENTS (bool): A flag indicating whether user
            components are allowed. Defaults to True.
        type (str): The type of the card. Defaults to "default".
    """

    ALLOW_USER_COMPONENTS = True

    type = "default"

    def __init__(self, options={"only_repr": True}, components=[], graph=None):
        """Initializes the DefaultCard with the provided options, components,
        and graph.

        Args:
            options (dict, optional): A dictionary of options. Defaults to
                {"only_repr": True}.
            components (list, optional): A list of components. Defaults to an
                empty list.
            graph (any, optional): The graph to use. Defaults to None.
        """
        self._only_repr = True
        self._graph = None if graph is None else transform_flow_graph(graph)
        if "only_repr" in options:
            self._only_repr = options["only_repr"]
        self._components = components

    # modified Defaultcard render function
    def render(self, task):
        """Renders the card with the provided task.

        Args:
            task (any): The task to render the card with.

        Returns:
            any: The rendered card.
        """
        # :param: task instead of metaflow.client.Task object task.pathspec
        # (string) is provided # NOQA
        RENDER_TEMPLATE = read_file(RENDER_TEMPLATE_PATH)  # NOQA: N806
        JS_DATA = read_file(JS_PATH)  # NOQA: N806
        CSS_DATA = read_file(CSS_PATH)  # NOQA: N806
        final_component_dict = dict(self._graph)
        final_component_dict = TaskInfoComponent(
            task,
            only_repr=self._only_repr,
            graph=self._graph,
            components=self._components,
        ).render()
        pt = self._get_mustache()
        data_dict = {
            "task_data": base64.b64encode(json.dumps(final_component_dict).encode("utf-8")).decode(
                "utf-8"
            ),
            "javascript": JS_DATA,
            "title": task,
            "css": CSS_DATA,
            "card_data_id": uuid.uuid4(),
        }
        return pt.render(RENDER_TEMPLATE, data_dict)


class TaskInfoComponent(TaskInfoComponent):
    """A custom TaskInfoComponent class for Metaflow.

    Properties:
        page_content (list): A list of MetaflowCardComponents going as task
            info.
        final_component (dict): The dictionary returned by the `render`
            function of this class.
    """

    def __init__(
        self,
        task,
        page_title="Task Info",
        only_repr=True,
        graph=None,
        components=[],
    ):
        """Initializes the TaskInfoComponent with the provided task, page
        title, representation flag, graph, and components.

        Args:
            task (any): The task to use.
            page_title (str, optional): The title of the page. Defaults to
                "Task Info".
            only_repr (bool, optional): A flag indicating whether to only use
                the representation. Defaults to True.
            graph (any, optional): The graph to use. Defaults to None.
            components (list, optional): A list of components. Defaults to an
                empty list.
        """
        self._task = task
        self._only_repr = only_repr
        self._graph = graph
        self._components = components
        self._page_title = page_title
        self.final_component = None
        self.page_component = None

    # modified TaskInfoComponent render function
    def render(self):
        """Renders the component and returns a dictionary of metadata and
        components.

        Returns:
            final_component_dict (dict): A dictionary of the form:
                dict(metadata={}, components=[]).
        """
        final_component_dict = {
            "metadata": {
                "metaflow_version": mf_version,
                "version": 1,
                "template": "defaultCardTemplate",
            },
            "components": [],
        }

        dag_component = SectionComponent(
            title="DAG", contents=[DagComponent(data=self._graph).render()]
        ).render()

        page_contents = []
        page_contents.append(dag_component)
        page_component = PageComponent(
            title=self._page_title,
            contents=page_contents,
        ).render()
        final_component_dict["components"].append(page_component)

        self.final_component = final_component_dict
        self.page_component = page_component

        return final_component_dict



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""
Mixin class for FL models. No default implementation.

Each framework will likely have its own baseclass implementation (e.g.
TensorflowTaskRunner) that uses this mixin.

You may copy use this file or the appropriate framework-specific base-class to
port your own models.
"""

import logging

logger = logging.getLogger(__name__)


class TaskRunner:
    """Federated Learning Task Runner Class.

    Attributes:
        data_loader: The data_loader object.
        tensor_dict_split_fn_kwargs (dict): Key word arguments for determining
            which parameters to hold out from aggregation.
        logger (logging.Logger): Logger object for logging events.
        opt_treatment (str): Treatment of current instance optimizer.
    """

    def __init__(self, data_loader, tensor_dict_split_fn_kwargs: dict = None, **kwargs):
        """Initializes the TaskRunner object.

        Args:
            data_loader: The data_loader object
            tensor_dict_split_fn_kwargs (dict, optional): Key word arguments
                for determining which parameters to hold out from aggregation.
                Default is None.
            **kwargs: Additional parameters to pass to the function.
        """
        self.data_loader = data_loader
        if self.data_loader:
            self.feature_shape = self.data_loader.get_feature_shape()
        else:
            self.feature_shape = None
        # TODO: Should this comment a path of the doc string?
        # key word arguments for determining which parameters to hold out from
        # aggregation.
        # If set to none, an empty dict will be passed, currently resulting in
        # the defaults:
        # be held out
        # holdout_tensor_names=[]                   # NOQA:E800
        # params with these names will be held out  # NOQA:E800
        # TODO: params are restored from protobufs as float32 numpy arrays, so
        # non-floats arrays and non-arrays are not currently supported for
        # passing to and from protobuf (and as a result for aggregation) - for
        # such params in current examples, aggregation does not make sense
        # anyway, but if this changes support should be added.
        if tensor_dict_split_fn_kwargs is None:
            tensor_dict_split_fn_kwargs = {}
        self.tensor_dict_split_fn_kwargs = tensor_dict_split_fn_kwargs

    def set_optimizer_treatment(self, opt_treatment):
        """Change the treatment of current instance optimizer.

        Args:
            opt_treatment (str): The optimizer treatment.

        Returns:
            None
        """
        self.opt_treatment = opt_treatment

    def get_data_loader(self):
        """Get the data_loader object.

        Serves up batches and provides info regarding data_loader.

        Returns:
            data_loader object
        """
        return self.data_loader

    def set_data_loader(self, data_loader):
        """Set data_loader object.

        Args:
            data_loader: data_loader object to set.

        Returns:
            None
        """
        if data_loader.get_feature_shape() != self.data_loader.get_feature_shape():
            raise ValueError("The data_loader feature shape is not compatible with model.")

        self.data_loader = data_loader

    def get_train_data_size(self):
        """Get the number of training examples.

        It will be used for weighted averaging in aggregation.

        Returns:
            int: The number of training examples.
        """
        return self.data_loader.get_train_data_size()

    def get_valid_data_size(self):
        """Get the number of examples.

        It will be used for weighted averaging in aggregation.

        Returns:
            int: The number of validation examples.
        """
        return self.data_loader.get_valid_data_size()

    def train_batches(self, num_batches=None, use_tqdm=False):
        """Perform the training for a specified number of batches.

        Is expected to perform draws randomly, without
        replacement until data is exhausted. Then data is replaced and
        shuffled and draws continue.

        Args:
            num_batches (int, optional): Number of batches to train. Default
                is None.
            use_tqdm (bool, optional): If True, use tqdm to print a progress
                bar. Default is False.

        Returns:
            dict: {<metric>: <value>}.
        """
        raise NotImplementedError

    def validate(self):
        """Run validation.

        Returns:
            dict: {<metric>: <value>}.
        """
        raise NotImplementedError

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """When running a task, a map of named tensorkeys must be provided to
        the function as dependencies.

        Args:
            func_name (str): The function name.
            **kwargs: Additional parameters to pass to the function.

        Returns:
            list: List of required TensorKey. (TensorKey(tensor_name, origin,
                round_number))
        """
        raise NotImplementedError

    def get_tensor_dict(self, with_opt_vars):
        """Get the weights.

        Args:
            with_opt_vars (bool): Specify if we also want to get the variables
                of the optimizer.

        Returns:
            dict: The weight dictionary {<tensor_name>: <value>}.
        """
        raise NotImplementedError

    def set_tensor_dict(self, tensor_dict, with_opt_vars):
        """Set the model weights with a tensor dictionary:
        {<tensor_name>: <value>}.

        Args:
            tensor_dict (dict): The model weights dictionary.
            with_opt_vars (bool): Specify if we also want to set the variables
                of the optimizer.

        Returns:
            None
        """
        raise NotImplementedError

    def reset_opt_vars(self):
        """Reinitialize the optimizer variables.

        Returns:
            None
        """
        raise NotImplementedError

    def initialize_globals(self):
        """Initialize all global variables.

        Returns:
            None
        """
        raise NotImplementedError

    def load_native(self, filepath, **kwargs):
        """Load model state from a filepath in ML-framework "native" format,
        e.g. PyTorch pickled models.

        May load from multiple files. Other filepaths may be derived from the
        passed filepath, or they may be in the kwargs.

        Args:
            filepath (str): Path to frame-work specific file to load.
                For frameworks that use multiple files, this string must be
                    used to derive the other filepaths.
            **kwargs: Additional parameters to pass to the function. For
                future-proofing.

        Returns:
            None
        """
        raise NotImplementedError

    def save_native(self, filepath, **kwargs):
        """Save model state in ML-framework "native" format, e.g. PyTorch
        pickled models.

        May save one file or multiple files, depending on the framework.

        Args:
            filepath (str): If framework stores a single file, this should be
                a single file path. Frameworks that store multiple files may
                need to derive the other paths from this path.
            **kwargs: Additional parameters to pass to the function. For
                future-proofing.

        Returns:
            None
        """
        raise NotImplementedError



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""
This file contains utility functions for Secure Aggregation for key operations.
"""

from Crypto.PublicKey import ECC  # nosec B413


def generate_key_pair(curve: str = "ed25519") -> tuple[str, str]:
    """
    Generates a public-private key pair for a specific curve.

    Args:
        curve (str, optional): The curve to use for generating the key pair.
            Defaults to 'ed25519'

    Returns:
        str: Private key in string.
        str: Public key in string.
    """
    # Generate private key.
    private_key = ECC.generate(curve=curve)
    # Generate public_key
    public_key = private_key.public_key().export_key(format="PEM")

    return private_key.export_key(format="PEM"), public_key


def generate_agreed_key(
    private_key: bytes,
    public_key: bytes,
) -> bytes:
    """
    Uses Diffie-Helman key agreement to generate an agreed key between a pair
    of public-private keys.

    Args:
        public_key (bytes): Public key to be used for key agreement.
        private_key (bytes): Private key to be used for key agreement.

    Returns:
        bytes: Agreed key between the two keys shared in args.
    """
    from Crypto.Hash import SHAKE128  # nosec B413

    # Key derivation function.
    def kdf(x):
        return SHAKE128.new(x).read(32)

    from Crypto.Protocol.DH import key_agreement  # nosec B413

    # Using Diffie-Hellman key agreement.
    key = key_agreement(
        static_priv=ECC.import_key(private_key), static_pub=ECC.import_key(public_key), kdf=kdf
    )

    return key



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Collaborator module."""

import os
import sys
from glob import glob
from logging import getLogger
from os import remove
from os.path import basename, isfile, join, splitext
from pathlib import Path
from shutil import copy, copytree, ignore_patterns, make_archive, unpack_archive
from tempfile import mkdtemp

from click import Path as ClickPath
from click import confirm, echo, group, option, pass_context, prompt, style
from yaml import FullLoader, dump, load

from openfl.cryptography.ca import sign_certificate
from openfl.cryptography.io import get_csr_hash, read_crt, read_csr, read_key, write_crt, write_key
from openfl.cryptography.participant import generate_csr
from openfl.federated import Plan
from openfl.federated.data.sources.data_sources_json_parser import DataSourcesJsonParser
from openfl.interface.cli_helper import CERT_DIR
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.utils import rmtree

logger = getLogger(__name__)


@group()
@pass_context
def collaborator(context):
    """Manage Federated Learning Collaborator."""
    context.obj["group"] = "service"


@collaborator.command(name="start")
@option(
    "-p",
    "--plan",
    required=False,
    help="Path to an FL plan.",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-d",
    "--data_config",
    required=False,
    help="The dataset shard configuration file.",
    default="plan/data.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-n",
    "--collaborator_name",
    required=True,
    help="The certified common name of the collaborator.",
)
def start_(plan, collaborator_name, data_config):
    """Starts a collaborator service."""
    if plan and is_directory_traversal(plan):
        echo("Federated learning plan path is out of the openfl workspace scope.")
        sys.exit(1)
    if data_config and is_directory_traversal(data_config):
        echo("The data set/shard configuration file path is out of the openfl workspace scope.")
        sys.exit(1)

    plan_obj = Plan.parse(
        plan_config_path=Path(plan).absolute(),
        data_config_path=Path(data_config).absolute(),
    )

    # TODO: Need to restructure data loader config file loader
    logger.info(f"Data paths: {plan_obj.cols_data_paths}")
    echo(f"Data = {plan_obj.cols_data_paths}")
    logger.info("🧿 Starting a Collaborator Service.")

    collaborator = plan_obj.get_collaborator(collaborator_name)
    collaborator.run()


@collaborator.command(name="ping")
@option(
    "-p",
    "--plan",
    required=False,
    help="Path to an FL plan.",
    default="plan/plan.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-d",
    "--data_config",
    required=False,
    help="The dataset shard configuration file.",
    default="plan/data.yaml",
    type=ClickPath(exists=True),
    show_default=True,
)
@option(
    "-n",
    "--collaborator_name",
    required=True,
    help="The certified common name of the collaborator.",
)
def ping_(plan, collaborator_name, data_config):
    """Ping the aggregator without starting any tasks."""

    if plan and is_directory_traversal(plan):
        echo("Federated learning plan path is out of the openfl workspace scope.")
        sys.exit(1)
    if data_config and is_directory_traversal(data_config):
        echo("The data set/shard configuration file path is out of the openfl workspace scope.")
        sys.exit(1)

    fl_plan = Plan.parse(
        plan_config_path=Path(plan).absolute(),
        data_config_path=Path(data_config).absolute(),
    )

    agg_addr = fl_plan.config["network"]["settings"]["agg_addr"]
    agg_port = fl_plan.config["network"]["settings"]["agg_port"]
    use_tls = fl_plan.config["network"]["settings"]["use_tls"]
    protocol = "TLS" if use_tls else "TCP"

    logger.info(
        f"🧿 Testing connectivity with the Aggregator at {agg_addr}:{agg_port} via {protocol}..."
    )
    fl_plan.get_collaborator(collaborator_name).ping()

    logger.info(f"The Aggregator is reachable at {agg_addr}:{agg_port}")
    logger.info(f"{protocol} connection established.")


@collaborator.command(name="create")
@option(
    "-n",
    "--collaborator_name",
    required=True,
    help="The certified common name of the collaborator.",
)
@option(
    "-d",
    "--data_path",
    help="The data path to be associated with the collaborator.",
)
@option("-s", "--silent", help="If set, skips manual confirmation.", is_flag=True)
def create_(collaborator_name, data_path, silent):
    """Registers given dataset path in the ``plan/data.yaml`` file."""
    create(collaborator_name, data_path, silent)


def create(collaborator_name, data_path, silent):
    """Creates a user for an experiment.

    Args:
        collaborator_name (str): The certified common name of the collaborator.
        data_path (str): The data path to be associated with the collaborator.
        silent (bool): Do not prompt.
    """
    if data_path and is_directory_traversal(data_path):
        echo("Data path is out of the openfl workspace scope.")
        sys.exit(1)

    common_name = f"{collaborator_name}".lower()

    # TODO: There should be some association with the plan made here as well
    register_data_path(common_name, data_path=data_path, silent=silent)


def register_data_path(collaborator_name, data_path=None, silent=False):
    """Register dataset path in the plan/data.yaml file.

    Args:
        collaborator_name (str): The collaborator whose data path to be
            defined.
        data_path (str, optional): Data path. Defaults to None.
        silent (bool, optional): Silent operation (don't prompt). Defaults to
            False.
    """

    if data_path and is_directory_traversal(data_path):
        echo("Data path is out of the openfl workspace scope.")
        sys.exit(1)

    # Ask for the data directory
    default_data_path = f"data/{collaborator_name}"
    if not silent and data_path is None:
        dir_path = prompt(
            "\nWhere is the data (or what is the rank)"
            " for collaborator " + style(f"{collaborator_name}", fg="green") + " ? ",
            default=default_data_path,
        )
    elif data_path is not None:
        dir_path = data_path
    else:
        # TODO: Need to figure out the default for this.
        dir_path = default_data_path

    # Read the data.yaml file
    d = {}
    data_yaml = "plan/data.yaml"
    separator = ","
    if isfile(data_yaml):
        with open(data_yaml, "r", encoding="utf-8") as f:
            for line in f:
                if separator in line:
                    key, val = line.split(separator, maxsplit=1)
                    d[key] = val.strip()

    d[collaborator_name] = dir_path

    # Write the data.yaml
    if isfile(data_yaml):
        with open(data_yaml, "w", encoding="utf-8") as f:
            for key, val in d.items():
                f.write(f"{key}{separator}{val}\n")


@collaborator.command(name="calchash")
@option(
    "-j",
    "--data_path",
    type=ClickPath(exists=True),
    help=(
        "Path to directory containing sources.json file defining the data sources of the dataset. "
        "This file should contain a JSON object with the data sources to be registered. For 'local'"
        " type, 'params' must include: 'path'. For 's3' type, 'params' must include: 'uri', "
        "'access_key_env_name', 'secret_key_env_name', 'secret_name', and optionally 'endpoint'."
    ),
)
def calchash(data_path):
    """Generates a hash for the dataset whose data sources are defined in `datasources.json`.

    The hash, saved as `hash.txt` in the same directory, can later
    be used in custom data loaders to validate and track dataset consistency.
    """

    if data_path and is_directory_traversal(data_path):
        logger.error("Data path is out of the openfl workspace scope.")
        sys.exit(1)
    if not os.path.isdir(data_path):
        logger.error("The data path must be a directory.")
        sys.exit(1)

    datasources_json_path = os.path.join(data_path, "datasources.json")
    if not os.path.isfile(datasources_json_path):
        logger.error(
            "The directory must contain a file named 'datasources.json' at the first level."
        )
        sys.exit(1)
    with open(datasources_json_path, "r", encoding="utf-8") as file:
        data = file.read()
    vds = DataSourcesJsonParser.parse(data)
    root_hash = vds.create_dataset_hash()
    hash_file_path = os.path.join(data_path, "hash.txt")
    with open(hash_file_path, "w", encoding="utf-8") as hash_file:
        hash_file.write(root_hash)

    logger.info(f"Dataset hash calculated and saved to {hash_file_path}. Hash: {root_hash}")


@collaborator.command(name="generate-cert-request")
@option(
    "-n",
    "--collaborator_name",
    required=True,
    help="The certified common name of the collaborator.",
)
@option("-s", "--silent", help="Do not prompt", is_flag=True)
@option(
    "-x",
    "--skip-package",
    help="If set, does not package the certificate signing request for export.",
    is_flag=True,
)
def generate_cert_request_(collaborator_name, silent, skip_package):
    """
    Generates collaborator certificate key-pair
    to be sent to aggregator for signing.
    """
    generate_cert_request(collaborator_name, silent, skip_package)


def generate_cert_request(collaborator_name, silent, skip_package):
    """Create collaborator certificate key pair.

    Then create a package with the CSR to send for signing.

    Args:
        collaborator_name (str): The certified common name of the collaborator.
        silent (bool): Do not prompt.
        skip_package (bool): Do not package the certificate signing request
            for export.
    """

    common_name = f"{collaborator_name}".lower()
    subject_alternative_name = f"DNS:{common_name}"
    file_name = f"col_{common_name}"

    echo(
        f"Creating COLLABORATOR certificate key pair with following settings: "
        f"CN={style(common_name, fg='red')},"
        f" SAN={style(subject_alternative_name, fg='red')}"
    )

    client_private_key, client_csr = generate_csr(common_name, server=False)

    (CERT_DIR / "client").mkdir(parents=True, exist_ok=True)

    echo("  Moving COLLABORATOR certificate to: " + style(f"{CERT_DIR}/{file_name}", fg="green"))

    # Print csr hash before writing csr to disk
    csr_hash = get_csr_hash(client_csr)
    echo("The CSR Hash " + style(f"{csr_hash}", fg="red"))

    # Write collaborator csr and key to disk
    write_crt(client_csr, CERT_DIR / "client" / f"{file_name}.csr")
    write_key(client_private_key, CERT_DIR / "client" / f"{file_name}.key")

    if not skip_package:
        archive_type = "zip"
        archive_name = f"col_{common_name}_to_agg_cert_request"
        archive_file_name = archive_name + "." + archive_type

        # Collaborator certificate signing request
        tmp_dir = join(mkdtemp(), "openfl", archive_name)

        ignore = ignore_patterns("__pycache__", "*.key", "*.srl", "*.pem")
        # Copy the current directory into the temporary directory
        copytree(f"{CERT_DIR}/client", tmp_dir, ignore=ignore)

        for f in glob(f"{tmp_dir}/*"):
            if common_name not in basename(f):
                remove(f)

        # Create Zip archive of directory
        make_archive(archive_name, archive_type, tmp_dir)
        rmtree(tmp_dir)

        echo(f"Archive {archive_file_name} with certificate signing request created")
        echo(
            "This file should be sent to the certificate authority"
            " (typically hosted by the aggregator) for signing"
        )


def find_certificate_name(file_name):
    """Parse the collaborator name.

    Args:
        file_name (str): The name of the collaborator in this federation.

    Returns:
        col_name (str): The collaborator name.
    """
    col_name = str(file_name).split(os.sep)[-1].split(".")[0][4:]
    return col_name


def register_collaborator(file_name):
    """Register the collaborator name in the cols.yaml list.

    Args:
        file_name (str): The name of the collaborator in this federation.
    """

    col_name = find_certificate_name(file_name)

    cols_file = Path("plan/cols.yaml").absolute()

    if not isfile(cols_file):
        cols_file.touch()
    with open(cols_file, "r", encoding="utf-8") as f:
        doc = load(f, Loader=FullLoader)

    if not doc:  # YAML is not correctly formatted
        doc = {}  # Create empty dictionary

    # List doesn't exist
    if "collaborators" not in doc.keys() or not doc["collaborators"]:
        doc["collaborators"] = []  # Create empty list

    if col_name in doc["collaborators"]:
        echo(
            "\nCollaborator "
            + style(f"{col_name}", fg="green")
            + " is already in the "
            + style(f"{cols_file}", fg="green")
        )

    else:
        doc["collaborators"].append(col_name)
        with open(cols_file, "w", encoding="utf-8") as f:
            dump(doc, f)

        echo(
            "\nRegistering "
            + style(f"{col_name}", fg="green")
            + " in "
            + style(f"{cols_file}", fg="green")
        )


@collaborator.command(name="certify")
@option("-n", "--collaborator_name", help="The certified common name of the collaborator.")
@option("-s", "--silent", help="Do not prompt", is_flag=True)
@option(
    "-r",
    "--request-pkg",
    type=ClickPath(exists=True),
    help="Path to a zip containing the certificate signing request for a collaborator.",
)
@option(
    "-i",
    "--import",
    "import_",
    type=ClickPath(exists=True),
    help="Path to an archive containing the collaborator's certificate (signed by the CA).",
)
def certify_(collaborator_name, silent, request_pkg, import_):
    """Certifies the collaborator certificate key-pair."""
    certify(collaborator_name, silent, request_pkg, import_)


def certify(collaborator_name, silent, request_pkg=None, import_=False):
    """Sign/certify collaborator certificate key pair.

    Args:
        collaborator_name (str): The certified common name of the collaborator.
        silent (bool): Do not prompt.
        request_pkg (str, optional): The archive containing the certificate
            signing request (*.zip) for a collaborator. Defaults to None.
        import_ (bool, optional): Import the archive containing the
            collaborator's certificate (signed by the CA). Defaults to False.
    """

    common_name = f"{collaborator_name}".lower()

    if not import_:
        if request_pkg:
            Path(f"{CERT_DIR}/client").mkdir(parents=True, exist_ok=True)
            unpack_archive(request_pkg, extract_dir=f"{CERT_DIR}/client")
            csr = glob(f"{CERT_DIR}/client/*.csr")[0]
        else:
            if collaborator_name is None:
                echo(
                    "collaborator_name can only be omitted if signing\n"
                    "a zipped request package.\n"
                    "\n"
                    "Example: fx collaborator certify --request-pkg "
                    "col_one_to_agg_cert_request.zip"
                )
                return
            csr = glob(f"{CERT_DIR}/client/col_{common_name}.csr")[0]
            copy(csr, CERT_DIR)
        cert_name = splitext(csr)[0]
        file_name = basename(cert_name)
        signing_key_path = "ca/signing-ca/private/signing-ca.key"
        signing_crt_path = "ca/signing-ca.crt"

        # Load CSR
        if not Path(f"{cert_name}.csr").exists():
            echo(
                style(
                    "Collaborator certificate signing request not found.",
                    fg="red",
                )
                + " Please run `fx collaborator generate-cert-request`"
                " to generate the certificate request."
            )

        csr, csr_hash = read_csr(f"{cert_name}.csr")

        # Load private signing key
        if not Path(CERT_DIR / signing_key_path).exists():
            echo(
                style("Signing key not found.", fg="red") + " Please run `fx workspace certify`"
                " to initialize the local certificate authority."
            )

        signing_key = read_key(CERT_DIR / signing_key_path)

        # Load signing cert
        if not Path(CERT_DIR / signing_crt_path).exists():
            echo(
                style("Signing certificate not found.", fg="red")
                + " Please run `fx workspace certify`"
                " to initialize the local certificate authority."
            )

        signing_crt = read_crt(CERT_DIR / signing_crt_path)

        echo(f"The CSR Hash for file {file_name}.csr is {csr_hash}")

        if silent:
            echo(
                "Signing COLLABORATOR certificate, "
                "Warning: manual check of certificate hashes is bypassed in silent mode."
            )
            signed_col_cert = sign_certificate(csr, signing_key, signing_crt.subject)
            write_crt(signed_col_cert, f"{cert_name}.crt")
            register_collaborator(CERT_DIR / "client" / f"{file_name}.crt")

        else:
            echo("Make sure the two hashes above are the same.")
            if confirm("Do you want to sign this certificate?"):
                echo(" Signing COLLABORATOR certificate")
                signed_col_cert = sign_certificate(csr, signing_key, signing_crt.subject)
                write_crt(signed_col_cert, f"{cert_name}.crt")
                register_collaborator(CERT_DIR / "client" / f"{file_name}.crt")

            else:
                echo(
                    style("Not signing certificate.", fg="red")
                    + " Please check with this collaborator to get the"
                    " correct certificate for this federation."
                )
                return

        if len(common_name) == 0:
            # If the collaborator name is provided, the collaborator and
            # certificate does not need to be exported
            return

        # Remove unneeded CSR
        remove(f"{cert_name}.csr")

        archive_type = "zip"
        archive_name = f"agg_to_{file_name}_signed_cert"

        # Collaborator certificate signing request
        tmp_dir = join(mkdtemp(), "openfl", archive_name)

        Path(f"{tmp_dir}/client").mkdir(parents=True, exist_ok=True)
        # Copy the signed cert to the temporary directory
        copy(f"{CERT_DIR}/client/{file_name}.crt", f"{tmp_dir}/client/")
        # Copy the CA certificate chain to the temporary directory
        copy(f"{CERT_DIR}/cert_chain.crt", tmp_dir)

        # Create Zip archive of directory
        make_archive(archive_name, archive_type, tmp_dir)
        rmtree(tmp_dir)

    else:
        _import_certificates(import_)


def _import_certificates(archive: str):
    # Copy the signed certificate and cert chain into PKI_DIR
    previous_crts = glob(f"{CERT_DIR}/client/*.crt")
    unpack_archive(archive, extract_dir=CERT_DIR)
    updated_crts = glob(f"{CERT_DIR}/client/*.crt")
    cert_difference = list(set(updated_crts) - set(previous_crts))
    if len(cert_difference) != 0:
        crt = basename(cert_difference[0])
        echo(f"Certificate {crt} installed to PKI directory")
    else:
        echo("Certificate updated in the PKI directory")



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""KCPipeline module."""

import copy as co
import gzip as gz

import numpy as np
from sklearn import cluster

from openfl.pipelines.pipeline import TransformationPipeline, Transformer


class KmeansTransformer(Transformer):
    """K-means transformer class for quantizing input data.

    This class is a transformer that uses the K-means method for quantization.

    Attributes:
        n_cluster (int): The number of clusters for the K-means.
        lossy (bool): Indicates if the transformer is lossy.
    """

    def __init__(self, n_cluster=6):
        """Initialize KmeansTransformer.

        Args:
            n_cluster (int, optional): The number of clusters for the K-means.
                Defaults to 6.
        """
        self.lossy = True
        self.n_cluster = n_cluster

    def forward(self, data, **kwargs):
        """Quantize data into n_cluster levels of values.

        Args:
            data: The data to be quantized.
            **kwargs: Variable arguments to pass.

        Returns:
            int_array: The quantized data.
            metadata: The metadata for the quantization.
        """
        metadata = {"int_list": list(data.shape)}
        # clustering
        k_means = cluster.KMeans(n_clusters=self.n_cluster, n_init=self.n_cluster)
        data = data.reshape((-1, 1))
        if data.shape[0] >= self.n_cluster:
            k_means = cluster.KMeans(n_clusters=self.n_cluster, n_init=self.n_cluster)
            k_means.fit(data)
            quantized_values = k_means.cluster_centers_.squeeze()
            indices = k_means.labels_
            quant_array = np.choose(indices, quantized_values)
        else:
            quant_array = data

        int_array, int2float_map = self._float_to_int(quant_array)
        metadata["int_to_float"] = int2float_map

        return int_array, metadata

    def backward(self, data, metadata, **kwargs):
        """Recover data array back to the original numerical type and the
        shape.

        Args:
            data: The flattened numpy array.
            metadata: The dictionary containing information for recovering to
                original data array.

        Returns:
            data: The numpy array with original numerical type and shape.
        """
        # convert back to float
        # TODO
        data = co.deepcopy(data)
        int2float_map = metadata["int_to_float"]
        for key in int2float_map:
            indices = data == key
            data[indices] = int2float_map[key]
        data_shape = list(metadata["int_list"])
        data = data.reshape(data_shape)
        return data

    @staticmethod
    def _float_to_int(np_array):
        """Create look-up table for conversion between floating and integer
        types.

        Args:
            np_array: A Numpy array.

        Returns:
            int_array: The input Numpy float array converted to an integer
                array.
            int_to_float_map: The dictionary mapping integers to floats.
        """
        flatten_array = np_array.reshape(-1)
        unique_value_array = np.unique(flatten_array)
        int_array = np.zeros(flatten_array.shape, dtype=np.int32)
        int_to_float_map = {}
        float_to_int_map = {}
        # create table
        for idx, u_value in enumerate(unique_value_array):
            int_to_float_map.update({idx: u_value})
            float_to_int_map.update({u_value: idx})
            # assign to the integer array
            indices = np.where(flatten_array == u_value)
            int_array[indices] = idx
        int_array = int_array.reshape(np_array.shape)
        return int_array, int_to_float_map


class GZIPTransformer(Transformer):
    """GZIP transformer class for losslessly compressing data.

    Attributes:
        lossy (bool): Indicates if the transformer is lossy.
    """

    def __init__(self):
        """Initialize GZIPTransformer."""
        self.lossy = False

    def forward(self, data, **kwargs):
        """Compress data into bytes.

        Args:
            data: A Numpy array.

        Returns:
            compressed_bytes_: The GZIP compressed data object.
            metadata: An empty dictionary.
        """
        bytes_ = data.astype(np.float32).tobytes()
        compressed_bytes_ = gz.compress(bytes_)
        metadata = {}
        return compressed_bytes_, metadata

    def backward(self, data, metadata, **kwargs):
        """Decompress data into numpy of float32.

        Args:
            data: Compressed GZIP data
            metadata: An empty dictionary.
            **kwargs: Additional parameters to pass to the function

        Returns:
            data: The decompressed data as a numpy array.
        """
        decompressed_bytes_ = gz.decompress(data)
        data = np.frombuffer(decompressed_bytes_, dtype=np.float32)
        return data


class KCPipeline(TransformationPipeline):
    """A pipeline class to compress data lossly using k-means and GZIP methods.

    Attributes:
        p (float): The amount of sparsity for compression.
        n_cluster (int): The number of K-mean clusters.
    """

    def __init__(self, p_sparsity=0.01, n_clusters=6, **kwargs):
        """Initialize a pipeline of transformers.

        Args:
            p_sparsity (float, optional): The amount of sparsity for
                compression. Defaults to 0.01.
            n_clusters (int, optional): The number of K-mean clusters.
                Defaults to 6.
            **kwargs: Additional keyword arguments.
        """
        # instantiate each transformer
        self.p = p_sparsity
        self.n_cluster = n_clusters
        transformers = [KmeansTransformer(self.n_cluster), GZIPTransformer()]
        super().__init__(transformers=transformers, **kwargs)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl path checks."""

import os
from pathlib import Path
from typing import Union


def is_directory_traversal(directory: Union[str, Path]) -> bool:
    """Check for directory traversal.

    This function checks if the provided directory is a subdirectory of the
    current working directory.
    It returns `True` if the directory is not a subdirectory (i.e., it is a
    directory traversal), and `False` otherwise.

    Args:
        directory (Union[str, Path]): The directory to check.

    Returns:
        bool: `True` if the directory is a directory traversal, `False`
            otherwise.
    """
    cwd = os.path.abspath(os.getcwd())
    requested_path = os.path.relpath(directory, start=cwd)
    requested_path = os.path.abspath(requested_path)
    common_prefix = os.path.commonprefix([requested_path, cwd])
    return common_prefix != cwd



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import tensorflow as tf
import keras

class CNNModel(keras.Model):
    """
    Custom Keras Model for a Convolutional Neural Network (CNN) with custom training and testing steps.
    This model showcase how to define a custom training and testing step for a Keras model with Tensorflow.
    Methods
    -------
    train_step(data)
        Performs a single training step, including forward pass, loss computation, gradient calculation,
        and weight updates. Also updates the metrics.
    test_step(data)
        Performs a single testing step, including forward pass, loss computation, and metric updates.
    """

    def train_step(self, data):
        """
        Perform a single training step.
        Args:
            data (tuple): A tuple containing the input data and labels. If the tuple has three elements,
                          it should be (x, y, sample_weight). Otherwise, it should be (x, y).
        Returns:
            dict: A dictionary mapping metric names to their current values. This includes the loss and
                  any other metrics configured in `compile()`.
        Notes:
            - The loss function and metrics are configured in the `compile()` method.
            - The optimizer is used to apply the computed gradients to the model's trainable variables.
        """

        # Unpack the data. Its structure depends on your model and
        # on what you pass to `fit()`.
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data

        with tf.GradientTape() as tape:
            y_pred = self(x, training=True)  # Forward pass
            # Compute the loss value.
            # The loss function is configured in `compile()`.
            loss = self.compute_loss(
                y=y,
                y_pred=y_pred,
                sample_weight=sample_weight,
            )

        # Compute gradients
        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        # Update weights
        self.optimizer.apply(gradients, trainable_vars)

        # Update the metrics.
        # Metrics are configured in `compile()`.
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(y, y_pred, sample_weight=sample_weight)

        # Return a dict mapping metric names to current value.
        # Note that it will include the loss (tracked in self.metrics).
        return {m.name: m.result() for m in self.metrics}

    def test_step(self, data):
        """
        Perform a single test step.
        Args:
            data (tuple): A tuple containing the input data (x) and the true labels (y).
        Returns:
            dict: A dictionary mapping metric names to their current values. This includes the loss and other metrics tracked in self.metrics.
        """

        # Unpack the data
        x, y = data
        # Compute predictions
        y_pred = self(x, training=False)
        # Updates the metrics tracking the loss
        loss = self.compute_loss(y=y, y_pred=y_pred)
        # Update the metrics.
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(y, y_pred)
        # Return a dict mapping metric names to current value.
        # Note that it will include the loss (tracked in self.metrics).
        return {m.name: m.result() for m in self.metrics}



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""CA module."""

import base64
import json
import os
import shutil
import signal
import subprocess  # nosec
import sys
import time
from logging import getLogger
from pathlib import Path
from subprocess import check_call  # nosec

from click import confirm

from openfl.utilities.ca.downloader import download_step_bin, download_step_ca_bin

logger = getLogger(__name__)

TOKEN_DELIMITER = "."
CA_STEP_CONFIG_DIR = Path("step_config")
CA_PKI_DIR = Path("cert")
CA_PASSWORD_FILE = Path("pass_file")
CA_CONFIG_JSON = Path("config/ca.json")


def get_token(name, ca_url, ca_path="."):
    """
    Create authentication token.

    Args:
        name (str): Common name for the following certificate
            (aggregator fqdn or collaborator name).
        ca_url (str): Full URL of the CA server.
        ca_path (str, optional): Path to the CA folder. Defaults to '.'.

    Returns:
        str: The generated authentication token.
    """
    ca_path = Path(ca_path)
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    pki_dir = ca_path / CA_PKI_DIR
    step_path, _ = get_ca_bin_paths(ca_path)
    if not step_path:
        raise Exception("Step-CA is not installed!\nRun `fx pki install` first")

    priv_json = step_config_dir / "secrets" / "priv.json"
    pass_file = pki_dir / CA_PASSWORD_FILE
    root_crt = step_config_dir / "certs" / "root_ca.crt"
    try:
        token = subprocess.check_output(
            [
                step_path,
                "ca",
                "token",
                name,
                "--key",
                priv_json,
                "--root",
                root_crt,
                "--password-file",
                pass_file,
                "--ca-url",
                ca_url,
            ]
        )
    except subprocess.CalledProcessError as exc:
        logger.error("Error code %s: %s", exc.returncode, exc.output)
        sys.exit(1)

    token = token.strip()
    token_b64 = base64.b64encode(token)

    with open(root_crt, mode="rb") as file:
        root_certificate_b = file.read()
    root_ca_b64 = base64.b64encode(root_certificate_b)

    return TOKEN_DELIMITER.join(
        [
            token_b64.decode("utf-8"),
            root_ca_b64.decode("utf-8"),
        ]
    )


def get_ca_bin_paths(ca_path):
    """Get the paths of the step binaries.

    Args:
        ca_path (str): Path to the CA directory.

    Returns:
        tuple: Paths to the step and step-ca binaries.
    """
    ca_path = Path(ca_path)
    step = None
    step_ca = None
    if (ca_path / "step").exists():
        dirs = os.listdir(ca_path / "step")
        for dir_ in dirs:
            if "step_" in dir_:
                step_executable = "step"
                if sys.platform == "win32":
                    step_executable = "step.exe"
                step = ca_path / "step" / dir_ / "bin" / step_executable
            if "step-ca" in dir_:
                step_ca_executable = "step-ca"
                if sys.platform == "win32":
                    step_ca_executable = "step-ca.exe"
                step_ca = ca_path / "step" / dir_ / "bin" / step_ca_executable
    return step, step_ca


def certify(name, cert_path: Path, token_with_cert, ca_path: Path):
    """Create a certificate for a given name.

    Args:
        name (str): Name for the certificate.
        cert_path (Path): Path to store the certificate.
        token_with_cert (str): Authentication token with certificate.
        ca_path (Path): Path to the CA directory.
    """
    os.makedirs(cert_path, exist_ok=True)

    token, root_certificate = token_with_cert.split(TOKEN_DELIMITER)
    token = base64.b64decode(token).decode("utf-8")
    root_certificate = base64.b64decode(root_certificate)

    step_path, _ = get_ca_bin_paths(ca_path)
    if not step_path:
        download_step_bin(prefix=ca_path)
        step_path, _ = get_ca_bin_paths(ca_path)
    if not step_path:
        raise Exception("Step-CA is not installed!\nRun `fx pki install` first")

    with open(f"{cert_path}/root_ca.crt", mode="wb") as file:
        file.write(root_certificate)
    check_call(
        [
            step_path,
            "ca",
            "certificate",
            name,
            f"{cert_path}/{name}.crt",
            f"{cert_path}/{name}.key",
            "--kty",
            "EC",
            "--curve",
            "P-384",
            "-f",
            "--token",
            token,
        ]
    )


def remove_ca(ca_path):
    """
    Remove the CA directory and kill the step-ca process.

    Args:
        ca_path (str): Path to the CA directory.
    """
    _check_kill_process("step-ca")
    shutil.rmtree(ca_path, ignore_errors=True)


def install(ca_path, ca_url, password):
    """Create a certificate authority for the federation.

    Args:
        ca_path (str): Path to the CA directory.
        ca_url (str): URL for the CA server. Like: 'host:port'
        password (str): Password for encrypting root private keys.
    """
    logger.info("Creating CA")

    ca_path = Path(ca_path)
    ca_path.mkdir(parents=True, exist_ok=True)
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    os.environ["STEPPATH"] = str(step_config_dir)
    step_path, step_ca_path = get_ca_bin_paths(ca_path)

    if not (step_path and step_ca_path and step_path.exists() and step_ca_path.exists()):
        download_step_bin(prefix=ca_path, confirmation=True)
        download_step_ca_bin(prefix=ca_path, confirmation=False)
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    if not step_config_dir.exists() or confirm(
        "CA exists, do you want to recreate it?", default=True
    ):
        _create_ca(ca_path, ca_url, password)
    _configure(step_config_dir)


def run_ca(step_ca, pass_file, ca_json):
    """
    Run the CA server.

    Args:
        step_ca (str): Path to the step-ca binary.
        pass_file (str): Path to the password file.
        ca_json (str): Path to the CA configuration JSON file.
    """
    if _check_kill_process("step-ca", confirmation=True):
        logger.info("Up CA server")
        check_call([step_ca, "--password-file", pass_file, ca_json])


def _check_kill_process(pstring, confirmation=False):
    """Kill a process by its name.

    Args:
        pstring (str): Name of the process.
        confirmation (bool, optional): If True, ask for confirmation before
            killing the process. Defaults to False.

    Returns:
        bool: True if the process was killed, False otherwise.
    """
    pids = []
    proc = subprocess.Popen(
        ["ps", "ax"],
        stdout=subprocess.PIPE,
    )
    grep_proc = subprocess.Popen(
        ["grep", pstring],
        stdin=proc.stdout,
        stdout=subprocess.PIPE,
    )
    proc.stdout.close()
    grep_proc_2 = subprocess.Popen(
        ["grep", "-v", "grep"],
        stdin=grep_proc.stdout,
        stdout=subprocess.PIPE,
    )
    grep_proc.stdout.close()
    text = grep_proc_2.communicate()[0].decode("utf-8")

    for line in text.splitlines():
        fields = line.split()
        pids.append(fields[0])

    if len(pids):
        if confirmation and not confirm("CA server is already running. Stop him?", default=True):
            return False
        for pid in pids:
            os.kill(int(pid), signal.SIGKILL)
        time.sleep(2)
    return True


def _create_ca(ca_path: Path, ca_url: str, password: str):
    """
    Create a certificate authority workspace.

    Args:
        ca_path (Path): Path to the CA directory.
        ca_url (str): URL for the CA server.
        password (str): Password for encrypting root private keys.
    """
    pki_dir = ca_path / CA_PKI_DIR
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR

    pki_dir.mkdir(parents=True, exist_ok=True)
    step_config_dir.mkdir(parents=True, exist_ok=True)

    with open(f"{pki_dir}/pass_file", "w", encoding="utf-8") as f:
        f.write(password)
    os.chmod(f"{pki_dir}/pass_file", 0o600)
    step_path, step_ca_path = get_ca_bin_paths(ca_path)
    if not (step_path and step_ca_path and step_path.exists() and step_ca_path.exists()):
        logger.error("Could not find step-ca binaries in the path specified")
        sys.exit(1)

    logger.info("Create CA Config")
    os.environ["STEPPATH"] = str(step_config_dir)
    shutil.rmtree(step_config_dir, ignore_errors=True)
    name = ca_url.split(":")[0]
    check_call(
        [
            step_path,
            "ca",
            "init",
            "--name",
            "name",
            "--dns",
            name,
            "--address",
            ca_url,
            "--provisioner",
            "prov",
            "--password-file",
            f"{pki_dir}/pass_file",
        ]
    )

    check_call([step_path, "ca", "provisioner", "remove", "prov", "--all"])

    check_call(
        [
            step_path,
            "crypto",
            "jwk",
            "create",
            f"{step_config_dir}/certs/pub.json",
            f"{step_config_dir}/secrets/priv.json",
            f"--password-file={pki_dir}/pass_file",
        ]
    )
    check_call(
        [step_path, "ca", "provisioner", "add", "provisioner", f"{step_config_dir}/certs/pub.json"]
    )


def _configure(step_config_dir):
    """Configure the certificate authority.

    Args:
        step_config_dir (str): Path to the step configuration directory.
    """
    conf_file = step_config_dir / CA_CONFIG_JSON
    with open(conf_file, "r+", encoding="utf-8") as f:
        data = json.load(f)
        data.setdefault("authority", {}).setdefault("claims", {})
        data["authority"]["claims"]["maxTLSCertDuration"] = f"{365 * 24}h"
        data["authority"]["claims"]["defaultTLSCertDuration"] = f"{365 * 24}h"
        data["authority"]["claims"]["maxUserSSHCertDuration"] = "24h"
        data["authority"]["claims"]["defaultUserSSHCertDuration"] = "24h"
        f.seek(0)
        json.dump(data, f, indent=4)
        f.truncate()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl common object types."""

from abc import ABCMeta
from collections import namedtuple

TensorKey = namedtuple("TensorKey", ["tensor_name", "origin", "round_number", "report", "tags"])
TaskResultKey = namedtuple("TaskResultKey", ["task_name", "owner", "round_number"])

Metric = namedtuple("Metric", ["name", "value"])
LocalTensor = namedtuple("LocalTensor", ["col_name", "tensor", "weight"])


class SingletonABCMeta(ABCMeta):
    """Metaclass for singleton instances.

    This metaclass ensures that only one instance of any class using it can be
    created.

    Attributes:
        _instances (dict): A dictionary mapping classes to their instances.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Use the singleton instance if it has already been created.

        Args:
            *args: Positional arguments to pass to the class constructor.
            **kwargs: Keyword arguments to pass to the class constructor.

        Returns:
            Any: The singleton instance of the class.
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonABCMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.runtime package Runtime class."""

from openfl.experimental.workflow.runtime.federated_runtime import FederatedRuntime
from openfl.experimental.workflow.runtime.local_runtime import LocalRuntime
from openfl.experimental.workflow.runtime.runtime import Runtime



# Copyright (C) 2020-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Setup script."""

from setuptools import Command, find_packages, setup
from setuptools.command.build_py import build_py


class BuildPackageProtos(Command):
    """Command to generate project *_pb2.py modules from proto files."""

    user_options = []

    def initialize_options(self):
        """Set default values for all the options that this command supports.

        Note that these defaults may be overridden by other
        commands, by the setup script, by config files, or by the
        command-line.  Thus, this is not the place to code dependencies
        between options; generally, 'initialize_options()' implementations
        are just a bunch of "self.foo = None" assignments.

        This method must be implemented by all command classes.
        """
        pass

    def finalize_options(self):
        """Set final values for all the options that this command supports.

        This is always called as late as possible, ie.  after any option
        assignments from the command-line or from other commands have been
        done.  Thus, this is the place to code option dependencies: if
        'foo' depends on 'bar', then it is safe to set 'foo' from 'bar' as
        long as 'foo' still has the same value it was assigned in
        'initialize_options()'.

        This method must be implemented by all command classes.
        """
        pass

    def run(self):
        """Build gRPC modules."""
        from grpc.tools import command
        command.build_package_protos('.')


class BuildPyGRPC(build_py):
    """Command for Python modules build."""

    def __init__(self, dist):
        """Create a sub-command to execute."""
        self.subcommand = BuildPackageProtos(dist)
        super().__init__(dist)

    def run(self):
        """Build Python and GRPC modules."""
        self.subcommand.run()
        super().run()


setup(
    name="openfl-nightly",
    version='1.9.0.dev202509200',
    author='OpenFL Team',
    description='Federated Learning for the Edge',
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type='text/markdown',
    url='https://github.com/securefederatedai/openfl',
    packages=find_packages(
        include=(
            "openfl",
            "openfl.*",
            "openfl-docker",
            "openfl-workspace",
            "openfl-tutorials",
        )
    ),
    include_package_data=True,
    setup_requires=['grpcio-tools>=1.56.2,<1.66.0'],  # ensure it is in-sync with `install_requires`
    install_requires=[
        'click',
        'psutil',
        'pyyaml',
        'rich',
        'dynaconf',
        'tqdm',
        'numpy',
        'requests>=2.32.0',
        'cloudpickle',
        'cryptography',
        'pandas',
        'scikit-learn',
        'flatten_json',
        'tensorboardX',
        'protobuf>=4.21,<6.0.0',
        'grpcio>=1.56.2,<1.66.0',
        'Flask==3.1.1',
    ],
    python_requires='>=3.10, <3.13',
    project_urls={
        'Bug Tracker': 'https://github.com/securefederatedai/openfl/issues',
        'Documentation': 'https://openfl.readthedocs.io/en/stable/',
        'Source Code': 'https://github.com/securefederatedai/openfl/tree/66796dd321ac3478133c458937f69fd37c801fd7',
    },
    classifiers=[
        'Environment :: Console',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: System :: Distributed Computing',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    entry_points={'console_scripts': ['fx=openfl.interface.cli:entry']},
    cmdclass={
        'build_py': BuildPyGRPC,
        'build_grpc': BuildPackageProtos,
    },
)



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""DirectorGRPCServer module."""

import asyncio
import logging
import uuid
from pathlib import Path
from typing import AsyncIterator, Optional, Union

import grpc
from grpc import aio, ssl_server_credentials

from openfl.experimental.workflow.protocols import director_pb2, director_pb2_grpc
from openfl.experimental.workflow.transport.grpc.exceptions import EnvoyNotFoundError
from openfl.experimental.workflow.transport.grpc.grpc_channel_options import channel_options
from openfl.protocols.utils import get_headers, proto_to_datastream

logger = logging.getLogger(__name__)

CLIENT_ID_DEFAULT = "__default__"


class DirectorGRPCServer(director_pb2_grpc.DirectorServicer):
    """
    Director transport class.

    This class implements a gRPC server for the Director, allowing it to
    communicate with envoys.

    Attributes:
        listen_uri (str): The URI that the server is serving on.
        tls (bool): Whether to use TLS for the connection.
        root_certificate (Optional[Union[Path, str]]): The path to the root certificate for the TLS
            connection.
        private_key (Optional[Union[Path, str]]): The path to the server's private key for the TLS
            connection.
        certificate (Optional[Union[Path, str]]): The path to the server's certificate for the TLS
            connection.
        server (grpc.Server): The gRPC server.
        root_dir (Path): Path to the root directory
        director (Director): The director that this server is serving.
    """

    def __init__(
        self,
        *,
        director_cls,
        tls: bool = True,
        root_certificate: Optional[Union[Path, str]] = None,
        private_key: Optional[Union[Path, str]] = None,
        certificate: Optional[Union[Path, str]] = None,
        listen_host: str = "[::]",
        listen_port: int = 50051,
        envoy_health_check_period: int = 0,
        director_config: Optional[Path] = None,
        **kwargs,
    ) -> None:
        """
        Initialize a DirectorGRPCServer object.

        Args:
            director_cls (Type[Director]): The class of the director.
            tls (bool, optional): Whether to use TLS for the connection.
                Defaults to True.
            root_certificate (Optional[Union[Path, str]]): The path
                to the root certificate for the TLS connection. Defaults to
                None.
            private_key (Optional[Union[Path, str]]): The path to
                the server's private key for the TLS connection. Defaults to
                None.
            certificate (Optional[Union[Path, str]]): The path to
                the server's certificate for the TLS connection. Defaults to
                None.
            listen_host (str, optional): The host to listen on. Defaults to
                '[::]'.
            listen_port (int, optional): The port to listen on. Defaults to
                50051.
            director_config (Optional[Path]): Path to director_config file
            **kwargs: Additional keyword arguments.
        """
        super().__init__()
        self.listen_uri = f"{listen_host}:{listen_port}"
        self.tls = tls
        self._fill_certs(root_certificate, private_key, certificate)
        self.server = None
        self.root_dir = Path.cwd()
        self.director = director_cls(
            tls=self.tls,
            root_certificate=self.root_certificate,
            private_key=self.private_key,
            certificate=self.certificate,
            envoy_health_check_period=envoy_health_check_period,
            director_config=director_config,
            **kwargs,
        )

    def _fill_certs(self, root_certificate, private_key, certificate) -> None:
        """Fill certificates.

        Args:
            root_certificate (Union[Path, str]): The path to the root
                certificate for the TLS connection.
            private_key (Union[Path, str]): The path to the server's private
                key for the TLS connection.
            certificate (Union[Path, str]): The path to the server's
                certificate for the TLS connection.
        """
        if self.tls:
            if not all([root_certificate, private_key, certificate]):
                raise ValueError("Incomplete certificates provided")

            self.root_certificate = Path(root_certificate).absolute()
            self.private_key = Path(private_key).absolute()
            self.certificate = Path(certificate).absolute()
        else:
            self.root_certificate = self.private_key = self.certificate = None

    def start(self) -> None:
        """Launch the DirectorGRPCServer"""
        loop = asyncio.get_event_loop()
        loop.create_task(self.director.start_experiment_execution_loop())
        loop.run_until_complete(self._run_server())

    async def _run_server(self) -> None:
        """Run the gRPC server."""
        self.server = aio.server(options=channel_options)
        director_pb2_grpc.add_DirectorServicer_to_server(self, self.server)

        if not self.tls:
            self.server.add_insecure_port(self.listen_uri)
        else:
            with open(self.private_key, "rb") as f:
                private_key_b = f.read()
            with open(self.certificate, "rb") as f:
                certificate_b = f.read()
            with open(self.root_certificate, "rb") as f:
                root_certificate_b = f.read()
            server_credentials = ssl_server_credentials(
                ((private_key_b, certificate_b),),
                root_certificates=root_certificate_b,
                require_client_auth=True,
            )
            self.server.add_secure_port(self.listen_uri, server_credentials)
        logger.info(f"Starting director server on {self.listen_uri}")
        await self.server.start()
        await self.server.wait_for_termination()

    def get_caller(self, context) -> str:
        """Get caller name from context.

        if tls == True: get caller name from auth_context
        if tls == False: get caller name from context header 'client_id'

        Args:
            context (grpc.ServicerContext): The context of the request.

        Returns:
            str: The name of the caller.
        """
        if self.tls:
            return context.auth_context()["x509_common_name"][0].decode("utf-8")
        headers = get_headers(context)
        client_id = headers.get("client_id", CLIENT_ID_DEFAULT)
        return client_id

    def EnvoyConnectionRequest(self, request, context) -> director_pb2.RequestAccepted:
        """Handles a connection request from an Envoy.

        Args:
            request (director_pb2.ConnectEnvoyRequest): The request from
                the envoy
            context (grpc.ServicerContext): The context of the request.

        Returns:
            director_pb2.RequestAccepted: Indicating if connection was accepted
        """
        logger.info(f"Envoy {request.envoy_name} is attempting to connect")
        is_accepted = self.director.ack_envoy_connection_request(request.envoy_name)
        if is_accepted:
            logger.info(f"Envoy {request.envoy_name} is connected")

        return director_pb2.RequestAccepted(accepted=is_accepted)

    async def UpdateEnvoyStatus(self, request, context) -> director_pb2.UpdateEnvoyStatusResponse:
        """Accept health check from envoy.

        Args:
            request (director_pb2.UpdateEnvoyStatusRequest): The request from
                the envoy.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            resp (director_pb2.UpdateEnvoyStatusResponse): The response to the
                request.
        """
        logger.debug("Updating envoy status: %s", request)
        try:
            health_check_period = self.director.update_envoy_status(
                envoy_name=request.name,
                is_experiment_running=request.is_experiment_running,
            )
        except EnvoyNotFoundError as exc:
            logger.error(exc)
            await context.abort(grpc.StatusCode.NOT_FOUND, str(exc))
        else:
            resp = director_pb2.UpdateEnvoyStatusResponse()
            resp.health_check_period.seconds = health_check_period

            return resp

    async def GetEnvoys(self, request, context) -> director_pb2.GetEnvoysResponse:
        """Get status of connected envoys.

        Args:
            request (director_pb2.GetEnvoysRequest): The request from
                the envoy.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            director_pb2.GetEnvoysResponse: The response to the request.
        """
        envoy_infos = self.director.get_envoys()
        envoy_statuses = []
        for envoy_name, envoy_info in envoy_infos.items():
            envoy_info_message = director_pb2.EnvoyInfo(
                envoy_name=envoy_name,
                is_online=envoy_info["is_online"],
                is_experiment_running=envoy_info["is_experiment_running"],
                experiment_name=envoy_info["experiment_name"],
            )
            envoy_info_message.valid_duration.seconds = envoy_info["valid_duration"]
            envoy_info_message.last_updated.seconds = int(envoy_info["last_updated"])

            envoy_statuses.append(envoy_info_message)

        return director_pb2.GetEnvoysResponse(envoy_infos=envoy_statuses)

    async def GetExperimentData(
        self, request, context
    ) -> AsyncIterator[director_pb2.ExperimentData]:
        """Receive experiment data.

        Args:
            request (director_pb2.GetExperimentDataRequest): The request from
                the collaborator.
            context (grpc.ServicerContext): The context of the request.

        Yields:
            director_pb2.ExperimentData: The experiment data.
        """
        data_file_path = self.director.get_experiment_data(request.experiment_name)
        max_buffer_size = 2 * 1024 * 1024
        with open(data_file_path, "rb") as df:
            while True:
                data = df.read(max_buffer_size)
                if len(data) == 0:
                    break
                yield director_pb2.ExperimentData(size=len(data), exp_data=data)

    async def WaitExperiment(self, request, context) -> director_pb2.WaitExperimentResponse:
        """Handles a request to wait for an experiment to be ready.

        Args:
            request (director_pb2.WaitExperimentRequest): The request from the
                collaborator.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            director_pb2.WaitExperimentResponse: The response to the request.
        """
        logger.debug(
            "Request WaitExperiment received from envoy %s",
            request.collaborator_name,
        )
        experiment_name = await self.director.wait_experiment(request.collaborator_name)
        logger.debug(
            "Experiment %s is ready for %s",
            experiment_name,
            request.collaborator_name,
        )

        return director_pb2.WaitExperimentResponse(experiment_name=experiment_name)

    async def SetNewExperiment(self, stream, context) -> director_pb2.SetNewExperimentResponse:
        """Request to set new experiment.

        Args:
            stream (grpc.aio._MultiThreadedRendezvous): The stream of
                experiment data.
            context (grpc.ServicerContext): The context of the request.

        Returns:
            director_pb2.SetNewExperimentResponse: The response to the request.
        """
        data_file_path = self.root_dir / str(uuid.uuid4())
        with open(data_file_path, "wb") as data_file:
            async for request in stream:
                if request.experiment_data.size == len(request.experiment_data.exp_data):
                    data_file.write(request.experiment_data.exp_data)
                else:
                    raise Exception("Could not register new experiment")

        caller = self.get_caller(context)

        is_accepted = await self.director.set_new_experiment(
            experiment_name=request.name,
            sender_name=caller,
            collaborator_names=request.collaborator_names,
            experiment_archive_path=data_file_path,
        )

        logger.info("Experiment %s registered", request.name)
        return director_pb2.SetNewExperimentResponse(status=is_accepted)

    async def GetFlowState(self, request, context) -> director_pb2.GetFlowStateResponse:
        """Get updated flow after experiment is finished.

        Args:
            request (director_pb2.GetFlowStatusRequest): The request from
                the experiment manager
            context (grpc.ServicerContext): The context of the request.

        Returns:
            director_pb2.GetFlowStateResponse: The response to the request.
        """
        status, flspec_obj = await self.director.get_flow_state()
        response = director_pb2.GetFlowStateResponse(
            completed=status,
            flspec_obj=flspec_obj,
        )
        for chunk in proto_to_datastream(response):
            await context.write(chunk)

    async def GetExperimentStdout(
        self, request, context
    ) -> AsyncIterator[director_pb2.GetExperimentStdoutResponse]:
        """
        Request to stream stdout from the aggregator to frontend.

        Args:
            request (director_pb2.GetExperimentStdoutRequest): The request from
                the experiment manager.
            context (grpc.ServicerContext): The context of the request.

        Yields:
            director_pb2.GetExperimentStdoutResponse: The metrics.
        """
        logger.info("Getting standard output for experiment: %s...", request.experiment_name)
        caller = self.get_caller(context)
        async for stdout_dict in self.director.stream_experiment_stdout(
            experiment_name=request.experiment_name, caller=caller
        ):
            if stdout_dict is None:
                await asyncio.sleep(1)
                continue
            yield director_pb2.GetExperimentStdoutResponse(**stdout_dict)



#!/usr/bin/env python
# Copyright (C) 2020-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""CLI module."""

import os
import re
import sys
import time
import warnings
from importlib import import_module
from pathlib import Path
from sys import argv, path

from click import (
    Group,
    argument,
    command,
    confirm,
    echo,
    group,
    open_file,
    option,
    pass_context,
    style,
)

import openfl
from openfl.utilities.logging import setup_logger


def disable_warnings():
    """Disables warnings."""

    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
    warnings.simplefilter(action="ignore", category=FutureWarning)
    warnings.simplefilter(action="ignore", category=UserWarning)


class CLI(Group):
    """CLI class."""

    def __init__(self, name=None, commands=None, **kwargs):
        """
        Initialize CLI object.

        Args:
            name (str, optional): Name of the CLI group. Defaults to None.
            commands (dict, optional): Commands for the CLI group. Defaults
                to None.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name, commands, **kwargs)
        self.commands = commands or {}

    def list_commands(self, ctx):
        """Display all available commands.

        Args:
            ctx (click.core.Context): Click context.

        Returns:
            dict: Available commands.
        """
        return self.commands

    def format_help(self, ctx, formatter):
        """Display user-friendly help.

        Args:
            ctx (click.core.Context): Click context.
            formatter (click.formatting.HelpFormatter): Click help formatter.
        """
        show_header()
        uses = [
            f"{ctx.command_path}",
            "[options]",
            style("[command]", fg="blue"),
            style("[subcommand]", fg="cyan"),
            "[args]",
        ]

        formatter.write(style("BASH COMPLETE ACTIVATION\n\n", bold=True, fg="bright_black"))
        formatter.write(
            "Run in terminal:\n"
            "   _FX_COMPLETE=bash_source fx > ~/.fx-autocomplete.sh\n"
            "   source ~/.fx-autocomplete.sh\n"
            "If ~/.fx-autocomplete.sh has already exist:\n"
            "   source ~/.fx-autocomplete.sh\n\n"
        )

        formatter.write(style("CORRECT USAGE\n\n", bold=True, fg="bright_black"))
        formatter.write(" ".join(uses) + "\n")

        opts = []
        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            if rv is not None:
                opts.append(rv)

        formatter.write(style("\nGLOBAL OPTIONS\n\n", bold=True, fg="bright_black"))
        formatter.write_dl(opts)

        cmds = []
        for cmd in self.list_commands(ctx):
            cmd = self.get_command(ctx, cmd)
            cmds.append((cmd.name, cmd, 0))

            for sub in cmd.list_commands(ctx):
                sub = cmd.get_command(ctx, sub)
                cmds.append((sub.name, sub, 1))

        formatter.write(style("\nAVAILABLE COMMANDS\n", bold=True, fg="bright_black"))

        for name, cmd, level in cmds:
            help_str = cmd.get_short_help_str()
            if level == 0:
                formatter.write(
                    f"\n{style(name, fg='blue', bold=True):<30} {style(help_str, bold=True)}" + "\n"
                )
                formatter.write("─" * 80 + "\n")
            if level == 1:
                formatter.write(
                    f"  {style('*', fg='green')} {style(name, fg='cyan'):<21} {help_str}" + "\n"
                )

    def invoke(self, ctx):
        if ctx.params.get("version"):
            echo(f"OpenFL version: {openfl.__version__}")
            ctx.exit()
        super().invoke(ctx)


@group(cls=CLI)
@option("-l", "--log-level", default="info", help="Logging verbosity level.")
@option("--no-warnings", is_flag=True, help="Disable third-party warnings.")
@option("-v", "--version", is_flag=True, help="Show version")
@pass_context
def cli(context, log_level, no_warnings, version):
    """Command-line Interface."""

    context.ensure_object(dict)
    context.obj["log_level"] = log_level
    context.obj["fail"] = False
    context.obj["script"] = argv[0]
    context.obj["arguments"] = argv[1:]

    if no_warnings:
        # Setup logging immediately to suppress unnecessary warnings on import
        # This will be overridden later with user selected debugging level
        disable_warnings()
    log_file = os.getenv("LOG_FILE")
    # Validate log_file with tighter restrictions
    if log_file:
        log_file = os.path.normpath(log_file)
        if (
            not re.match(r"^logs/[\w\-.]+$", log_file)
            or ".." in log_file
            or log_file.startswith("/")
        ):
            raise ValueError("Invalid log file path")
        # Ensure the log file is in the 'logs' directory
        allowed_directory = Path("logs").resolve()
        full_path = (allowed_directory / log_file).resolve()
        if not str(full_path).startswith(str(allowed_directory)):
            raise ValueError("Log file path is not allowed")
    setup_logger(log_level, log_file)
    sys.stdout.reconfigure(encoding="utf-8")


@cli.result_callback()
@pass_context
def end(context, result, **kwargs):
    """
    Print the result of the operation.

    Args:
        context (click.core.Context): Click context.
        result: Result of the operation.
        **kwargs: Arbitrary keyword arguments.
    """
    if context.obj["fail"]:
        echo("\n ❌ :(")
    else:
        echo("\n ✔️ OK")


@command(name="help")
@pass_context
@argument("subcommand", required=False)
def help_(context, subcommand):
    """Display help.

    Args:
        context (click.core.Context): Click context.
        subcommand (str, optional): Subcommand to display help for. Defaults
            to None.
    """
    pass


def error_handler(error):
    """
    Handle the error.

    Args:
        error (Exception): Error to handle.
    """
    if "cannot import" in str(error):
        if "TensorFlow" in str(error):
            echo(
                style("EXCEPTION", fg="red", bold=True)
                + " : "
                + style(
                    "Tensorflow must be installed prior to running this command",
                    fg="red",
                )
            )
        if "PyTorch" in str(error):
            echo(
                style("EXCEPTION", fg="red", bold=True)
                + " : "
                + style(
                    "Torch must be installed prior to running this command",
                    fg="red",
                )
            )
    echo(style("EXCEPTION", fg="red", bold=True) + " : " + style(f"{error}", fg="red"))
    raise error


def review_plan_callback(file_name, file_path):
    """
    Review plan callback for Director and Envoy.

    Args:
        file_name (str): Name of the file to review.
        file_path (str): Path of the file to review.

    Returns:
        bool: True if the file is accepted, False otherwise.
    """
    echo(
        style(
            f"Please review the contents of {file_name} before proceeding...",
            fg="green",
            bold=True,
        )
    )
    # Wait for users to read the question before flashing the contents of the file.
    time.sleep(3)

    with open_file(file_path, "r") as f:
        echo(f.read())

    if confirm(style(f"Do you want to accept the {file_name}?", fg="green", bold=True)):
        echo(style(f"{file_name} accepted!", fg="green", bold=True))
        return True
    else:
        echo(style(f"EXCEPTION: {file_name} rejected!", fg="red", bold=True))
        return False


def show_header():
    """Show header."""

    banner = "OpenFL - Open Federated Learning"

    experimental = (
        Path(os.path.expanduser("~")).resolve().joinpath(".openfl", "experimental").resolve()
    )

    if os.path.exists(experimental):
        banner = "OpenFL - Open Federated Learning (Experimental)"

    echo(style(f"{banner:<80}", bold=True, bg="bright_blue"))
    echo()


def entry():
    """Entry point of the Command-Line Interface."""

    experimental = (
        Path(os.path.expanduser("~")).resolve().joinpath(".openfl", "experimental").resolve()
    )

    root = Path(__file__).parent.resolve()

    if experimental.exists():
        root = root.parent.joinpath("experimental", "workflow", "interface", "cli").resolve()

    work = Path.cwd().resolve()
    path.append(str(root))
    path.insert(0, str(work))

    for module in root.glob("*.py"):  # load command modules
        package = module.parent
        module = module.name.split(".")[0]

        if module.count("__init__") or module.count("cli"):
            continue

        command_group = import_module(module, package)

        cli.add_command(command_group.__getattribute__(module))

    try:
        cli(max_content_width=120)
    except Exception as e:
        error_handler(e)


if __name__ == "__main__":
    entry()



# Copyright (C) 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Envoy CLI."""

import logging
import sys
from pathlib import Path

import click
from click import Path as ClickPath
from click import group, option, pass_context
from dynaconf import Validator

from openfl.experimental.workflow.component.envoy import Envoy
from openfl.utilities import is_fqdn, merge_configs
from openfl.utilities.path_check import is_directory_traversal

logger = logging.getLogger(__name__)


@group()
@pass_context
def envoy(context):
    """Manage Federated Learning Envoy.

    Args:
        context (click.core.Context): Click context.
    """
    context.obj["group"] = "envoy"


@envoy.command(name="start")
@option("-n", "--envoy_name", required=True, help="Current shard name")
@option(
    "-c",
    "--envoy-config-path",
    default="envoy_config.yaml",
    help="The envoy config path",
    type=ClickPath(exists=True),
)
@option(
    "--tls/--disable-tls",
    default=True,
    is_flag=True,
    help="Use TLS or not (By default TLS is enabled)",
)
@option(
    "-rc",
    "--root-cert-path",
    "root_certificate",
    default=None,
    help="Path to a root CA cert",
    type=ClickPath(exists=True),
)
@option(
    "-pk",
    "--private-key-path",
    "private_key",
    default=None,
    help="Path to a private key",
    type=ClickPath(exists=True),
)
@option(
    "-oc",
    "--public-cert-path",
    "certificate",
    default=None,
    help="Path to a signed certificate",
    type=ClickPath(exists=True),
)
def start_(
    envoy_name,
    tls,
    envoy_config_path,
    root_certificate,
    private_key,
    certificate,
):
    """Start the Envoy.

    Args:
        envoy_name (str): Name of the Envoy.
        tls (bool): Use TLS or not.
        envoy_config_path (str): The envoy config path.
        root_certificate (str): Path to a root CA cert.
        private_key (str): Path to a private key.
        certificate (str): Path to a signed certificate.
    """

    logger.info("🧿 Starting the Envoy.")
    if is_directory_traversal(envoy_config_path):
        click.echo("The envoy config path is out of the openfl workspace scope.")
        sys.exit(1)

    config = merge_configs(
        settings_files=envoy_config_path,
        overwrite_dict={
            "root_certificate": root_certificate,
            "private_key": private_key,
            "certificate": certificate,
        },
        validators=[
            Validator(
                "settings",
                must_exist=True,
                messages={"must_exist_true": "Missing 'settings' in Envoy Configuration file."},
            ),
            Validator(
                "settings.director_host",
                must_exist=True,
                condition=lambda x: bool(x) and is_fqdn(x),
                messages={
                    "must_exist_true": "Missing 'director_host' in Envoy Configuration file",
                    "condition": "Invalid 'director_host' in Envoy Configuration file."
                    "Must be a valid FQDN",
                },
            ),
            Validator(
                "settings.director_port",
                must_exist=True,
                condition=lambda value: isinstance(value, int) and 1024 <= value <= 65535,
                messages={
                    "must_exist_true": "Missing 'director_port' in Envoy Configuration file",
                    "condition": "Invalid 'director_port' in Envoy Configuration file"
                    "Must be an integer between 1024 & 65535",
                },
            ),
            Validator("params.install_requirements", default=True),
        ],
    )

    # Parse envoy parameters
    envoy_params = config.get("params", {})
    if envoy_params:
        install_requirements = envoy_params["install_requirements"]
    else:
        install_requirements = False

    if config.root_certificate:
        config.root_certificate = Path(config.root_certificate).absolute()
    if config.private_key:
        config.private_key = Path(config.private_key).absolute()
    if config.certificate:
        config.certificate = Path(config.certificate).absolute()

    envoy = Envoy(
        envoy_name=envoy_name,
        director_host=config.settings.director_host,
        director_port=config.settings.director_port,
        envoy_config=Path(envoy_config_path).absolute(),
        root_certificate=config.root_certificate,
        private_key=config.private_key,
        certificate=config.certificate,
        tls=tls,
        install_requirements=install_requirements,
    )

    envoy.start()
